import numpy as np
import pandas as pd

def slicer(vals, nlevels, acctpos, expand_with=''):
    expand_with = expand_with if expand_with else slice(None, None, None)
    
    slicearr = np.repeat(None, nlevels)
    slicearr[acctpos] = [vals] if isinstance(vals, str) else vals
    
    if nlevels > 1:
        slicearr[np.arange(nlevels) != acctpos] = expand_with
    
    return tuple(slicearr)

### ACCOUNT OBJECTS and MAIN LIST ###
class Account:
    ALLOWED_TYPES = ['Account', 'Prior', 'Equal', 'Sum', 'Sub', 'Metric']
    CORE = ALLOWED_TYPES[:-1]
    TABS = ALLOWED_TYPES[1:-1]

    def __init__(self, 
        name:str, 
        statclass:str='',  
        type_:str='Account',
        feeders=None, 
        has_levels=True,
        hide=False,
        short:str=''
    ):
        if type_ not in self.ALLOWED_TYPES:
            raise ValueError(f'`type_` not supported. Require one of {self.ALLOWED_TYPES}')
        
        self.name = name
        self.statclass = statclass
        self.type = type_
        self.feeders = feeders
        self.has_levels = has_levels if self.statclass != 'BSheet' else False
        self.hide = hide
        self._short = short

    def __repr__(self):
        return f'Account: {self.name} {self.statclass} {self.type} {self.feeders}'

    @property
    def short(self):
        if self._short:
            return self._short
        if len(self.name.split(' ')) > 1:
            return ''.join(s[0].lower() for s in self.name.split(' '))
        else:
            return self.name.lower()

    @property
    def show(self):
        return not self.hide

    @property
    def is_sum(self):
        return self.type == 'Sum'

    @property
    def is_sub(self):
        return self.type == 'Sub'

    @property
    def is_divide(self):
        return self.type == 'Divide'

    @property
    def is_metric(self):
        return self.type == 'Metric'

    @property
    def is_prior(self):
        return self.type == 'Prior'

    @property
    def is_equal(self):
        return self.type == 'Equal'

    # Append Func and Calc Funcs #
    def append(self, stat):
        if self.name in stat.index.get_level_values('Account'):
            stat = stat.drop(self.name, level='Account')

        metric = self.calc(stat)
        if not self.has_levels:
            metric = metric if metric.ndim > 1 else metric.reshape(1,-1)
            newindex = stat.index.to_frame().iloc[:1, :]
            newindex.Segment = '---'
        else:
            acct_slice = slicer(self.feeders[0], stat.index.nlevels, self._acctpos(stat))
            newindex = stat.loc[acct_slice].index.to_frame()

        newindex.Account = self.name
        newindex = pd.MultiIndex.from_frame(newindex.reset_index(drop=True))
        newframe = pd.DataFrame(metric, index=newindex, columns=stat.columns)
        
        stat = stat.append(newframe)

        return stat

    def calc(self, *args, **kwargs):
        if self.is_sum:
            res = self.sum(*args, **kwargs)
        elif self.is_sub:
            res = self.subtract(*args, **kwargs)
        elif self.is_prior:
            res = self.calc_prior(*args, **kwargs)
        elif self.is_equal:
            res = self.equal(*args, **kwargs)
        else:
            res = None

        return res
    
    def _acctpos(self, stat):
        return stat.index.names.index('Account')

    def sum(self, stat):
        add_slice = slicer(self.feeders, stat.index.nlevels, self._acctpos(stat))
        add_ = stat.loc[add_slice]
        
        if not self.has_levels:
            loop = stat.index.nlevels if stat.index.nlevels > 1 else 2
            for i in range(loop - 1):
                add_ = add_.sum()

        return add_.values

    def subtract(self, stat):
        add_slice = slicer(self.feeders[0], stat.index.nlevels, self._acctpos(stat))
        sub_slice = slicer(self.feeders[1:], stat.index.nlevels, self._acctpos(stat))
        
        add_ = stat.loc[add_slice]
        sub_ = stat.loc[sub_slice]

        if not self.has_levels:
            loop = stat.index.nlevels if stat.index.nlevels > 1 else 2
            for i in range(stat.index.nlevels - 1):
                add_ = add_.sum()
                sub_ = sub_.sum()

        return add_.values - sub_.values

    def divide(self, stat):
        num_slice = slicer(self.feeders[0], stat.index.nlevels, self._acctpos(stat))
        denom_slice = slicer(self.feeders[1], stat.index.nlevels, self._acctpos(stat))
        
        num = stat.loc[num_slice]
        denom = stat.loc[denom_slice]

        if not self.has_levels:
            loop = stat.index.nlevels if stat.index.nlevels > 1 else 2
            for i in range(stat.index.nlevels - 1):
                num = num.sum()
                denom = denom.sum()

        return np.divide(num.values, denom.values)

    def calc_prior(self, stat):
        return stat.loc[self.feeders[0], :].shift(1, axis=1).fillna(0).values

    def equal(self, stat):
        return stat.loc[self.feeders[0], :].fillna(0).values

### ACCOUNT CHILDS ###

class Metric(Account):
    def __init__(self, *args, **kwargs):
        self.type = 'Metric'
        super().__init__(*args, **kwargs)

class TotalRevenue(Metric):
    def __init__(self, *args, **kwargs):
        super().__init__(
            'Total Revenue', 
            *args, 
            feeders=['Revenue'], 
            has_levels=False, 
            hide=True,
            **kwargs
        )

    def calc(self, stat):
        return self.sum(stat)

class TotalGrossProfit(Account):
    def __init__(self, *args, **kwargs):
        super().__init__(
            'Total Gross Profit', 
            feeders=['Gross Profit'], 
            has_levels=False,
            hide=True,
            *args, 
            **kwargs
        )

    def calc(self, stat):
        return self.sum(stat)

class GrossMargin(Metric):
    def __init__(self, *args, **kwargs):
        super().__init__('Gross Margin', *args, **kwargs)

    def calc(self):
        pass

class EBIT(Metric):
    def __init__(self, *args, **kwargs):
        super().__init__(
            'EBIT', 
            *args,
            feeders=['EBITDA', 'Amortization'],
            has_levels=False,
            hide=True,
            **kwargs,
        )

    def calc(self, stat):
        return self.subtract(stat)

class EBT(Metric):
    def __init__(self, *args, **kwargs):
        super().__init__(
            'EBT', 
            *args, 
            feeders=['EBIT', 'Interest'],
            has_levels=False,
            hide=True,
            **kwargs
        )

    def calc(self, stat):
        return self.subtract(stat)

class StockSaleProceeds(Account):
    FULL_NAME = 'Stock Sale Proceeds'

    def __init__(self, *args, **kwargs):
        super().__init__(self.FULL_NAME, *args, **kwargs)

class DeltaWoCap(Metric):
    FULL_NAME = 'Change in Non-Cash Working Capital'

    def __init__(self, *args, **kwargs):
        super().__init__(self.FULL_NAME, short='wocap', *args, **kwargs)

    def calc(self, stat):
        from .finstat import BSheet

        delta_wocass = stat.loc[BSheet.WOCASS[1:].names.tolist()].sum().diff().fillna(0).values
        delta_woclia = stat.loc[BSheet.WOCLIA[1:].names.tolist()].sum().diff().fillna(0).values
        
        return delta_woclia - delta_wocass

class PurchasePPE(Account):
    FULL_NAME = 'Purchases of PPE'

    def __init__(self, *args, **kwargs):
        super().__init__(self.FULL_NAME, short='ppe', *args, **kwargs)

class Ratio(Metric):
    def __init__(self, *args, **kwargs):
        super().__init__(
            self.FULL_NAME, 
            *args, 
            has_levels=False,
            hide=True,
            **kwargs
        )
    
    def calc(self, stat):
        return self.divide(stat)    

class WagesPerRev(Ratio):
    FULL_NAME = 'Salaries per Rev'

    def __init__(self, *args, **kwargs):
        super().__init__(
            *args, 
            short='sal_per_rev',
            feeders=['Salaries and Payroll Taxes', 'Total Revenue'],
            **kwargs)

class OthOpexPerRev(Ratio):
    FULL_NAME = 'Other Opex per Rev'

    def __init__(self, *args, **kwargs):
        super().__init__(
            *args, 
            short='oth_opex_per_rev',
            feeders=['Other Expenses', 'Total Revenue'],
            **kwargs)

class TOpexPerRev(Ratio):
    FULL_NAME = 'Total Opex per Rev'

    def __init__(self, *args, **kwargs):
        super().__init__(
            *args, 
            short='topex_per_rev',
            feeders=['Total Operating Expenses', 'Total Revenue'],
            **kwargs
        )

class WagesPerTOpex(Ratio):
    FULL_NAME = 'Salary-Opex Mix'

    def __init__(self, *args, **kwargs):
        super().__init__(
            *args, 
            short='sal_v_topex',
            feeders=['Salaries and Payroll Taxes', 'Total Operating Expenses'],
            **kwargs)

class OthOpexPerTOpex(Ratio):
    FULL_NAME = 'Other-Opex Mix'

    def __init__(self, *args, **kwargs):
        super().__init__(
            *args, 
            short='oth_v_topex',
            feeders=['Other Expenses', 'Total Operating Expenses'],
            **kwargs)

### ACCOUNT CONTAINERS ###
class Accounts(np.ndarray, Account):
    def __new__(cls, accounts:list):
        obj = np.asarray(accounts).view(cls)
        return obj

    def __array_finalize__(self, obj):
        if obj is None: return

    def __init__(self, accounts:list):
        for acct in accounts:
            setattr(self, acct.name, acct)

    def __repr__(self):
        return f'Accounts Type: {self}'

    def __add__(self, accounts):
        return Accounts([acct for acct in self] + [acct for acct in accounts])

    def get(self, acct_name):
        if acct_name in self.__dict__:
            return self.__dict__[acct_name]
        elif acct_name in self.shortkey:
            return self.__dict__[self.shortkey[acct_name]]

    def get_from_list(self, acct_names):
        return Accounts([self.get(acct_name) for acct_name in acct_names])

    # Properties that return another Accounts instance
    @property
    def accounts(self):
        return Accounts([item for item in self if item.type == 'Account'])

    @property
    def core(self):
        return Accounts([item for item in self if item.type in self.CORE])

    @property
    def tabs(self):
        return Accounts([acct for acct in self if acct.type in self.TABS])

    @property
    def metrics(self):
        return Accounts([item for item in self if item.type == 'Metric'])

    # Properties that return list
    @property
    def names(self):
        return np.array([item.name for item in self])

    # Properties that return list
    @property
    def shorts(self):
        return np.array([item.short for item in self])

    @property
    def shortkey(self):
        return {k:v for k,v in zip(self.shorts, self.names)}

    def by_stat(self, statclass):
        return Accounts([item for item in self if item.statclass in statclass])

class BSheetAccounts(Accounts):
    def __new__(cls, accounts:list):
        for acct in accounts:
            acct.statclass = 'BSheet'
            acct.has_levels = False

        obj = np.asarray(accounts).view(cls)

        return obj

class EQStatAccounts(Accounts):
    def __new__(cls, accounts:list):
        for acct in accounts:
            acct.statclass = 'EQStat'
            acct.has_levels = False

        obj = np.asarray(accounts).view(cls)

        return obj

class CFlowAccounts(Accounts):
    def __new__(cls, accounts:list):
        for acct in accounts:
            acct.statclass = 'CFlow'
            acct.has_levels = False

        obj = np.asarray(accounts).view(cls)

        return obj

class Capital(Metric):
    FULL_NAME = 'Capital'

    def __init__(self, *args, **kwargs):
        super().__init__(
            self.FULL_NAME,
            feeders=['Indebtedness', "Shareholders' Equity"],
            has_levels=False, 
            short='cap',
            hide=True,
            *args,
            **kwargs
        )

    def calc(self, stat):
        return self.sum(stat)    

class Debt2Cap(Metric):
    FULL_NAME = 'Debt-to-Capital'
    def __init__(self, *args, **kwargs):
        super().__init__(
            self.FULL_NAME,
            *args, 
            feeders=['Indebtedness', 'Capital'], 
            has_levels=False, 
            short='d2cap',
            hide=True,
            **kwargs
        )

    def calc(self, stat):
        return self.divide(stat)

class CurrentRatio(Metric):
    FULL_NAME = 'Current Ratio'
    def __init__(self, *args, **kwargs):
        super().__init__( 
            self.FULL_NAME,
            *args, 
            feeders=['Current Assets', 'Current Liabilities'], 
            has_levels=False, 
            short='crat',
            hide=True,
            **kwargs
        )

    def calc(self, stat):
        return self.divide(stat)



ISTAT_ACCOUNTS = Accounts([
    Account('Revenue', 'RevStat'),
    TotalRevenue('RevStat'), 
    Account('COGS', 'RevStat'), 
    Account('Gross Profit', 'RevStat', type_='Sub', feeders=['Revenue', 'COGS']), 
    TotalGrossProfit('IStat'), 
    Account('Gross Margin', 'RevStat', type_='Metric'),
    Account('Salaries and Payroll Taxes', 'WageStat'),
    Account('Benefits', 'WageStat'),
    Account('Other Expenses', 'OpexStat'),
    Account('Total Operating Expenses', 'IStat', type_='Sum', feeders=['Salaries and Payroll Taxes', 'Benefits', 'Other Expenses'], has_levels=False),
    Account('EBITDA', 'IStat', type_='Sub', feeders=['Gross Profit', 'Total Operating Expenses'], has_levels=False),
    Account('Amortization', 'AmortSched'),
    EBIT('IStat'),
    Account('Interest', 'IStat'),
    EBT('IStat'),
    Account('Taxes', 'IStat'),
    Account('Net Income', 'IStat', type_='Sub', feeders=['EBITDA', 'Amortization', 'Interest', 'Taxes'], has_levels=False),
    # WagesPerRev('IStat'),
    # OthOpexPerRev('IStat'),
    # TOpexPerRev('IStat'),
    # WagesPerTOpex('IStat'),
    # OthOpexPerTOpex('IStat'),
])

BSHEET_ACCOUNTS = BSheetAccounts([
    Account('Cash'),
    Account('Accounts Receivable'),
    Account('Inventory', short='inv'),
    Account('Current Assets', type_='Sum', feeders=['Cash', 'Accounts Receivable', 'Inventory']),
    Account('Capital Assets', short='capass'),
    Account('Tax Carry-Forward', short='taxfwd'),
    Account('Total Assets', type_='Sum', feeders=['Current Assets', 'Capital Assets', 'Tax Carry-Forward']),
    Account('Indebtedness', short='indebt'),
    Account('Accounts Payable'),
    Account('Current Liabilities', type_='Sum', feeders=['Indebtedness', 'Accounts Payable']),
    Account('Long-Term Debt', short='ltd', hide=True),
    Account('Total Liabilities', type_='Sum', feeders=['Current Liabilities', 'Long-Term Debt']),
    Account('Retained Earnings'),
    Account('Equity Capital'),
    Account("Shareholders' Equity", type_='Sum', feeders=['Retained Earnings', 'Equity Capital']),
    Account('Liabilities and Equity', type_='Sum', feeders=['Total Liabilities', "Shareholders' Equity"]),
    Account('bal',  'BSheet', type_='Sub', feeders=['Total Assets', 'Liabilities and Equity']),
    Account('Shadow Cash', hide=True),
    Account('Interest', hide=True),
    Account('Taxes', hide=True),
    Account('Cash Taxes', hide=True),
    Account('Non-Cash Taxes', hide=True),
    Account('Net Income', hide=True),
    DeltaWoCap(hide=True),
    PurchasePPE(hide=True),
    StockSaleProceeds(hide=True),
    CurrentRatio(),
    Capital(),
    Debt2Cap(),
])

EQ_ACCOUNTS = EQStatAccounts([
    Account('Retained Earnings - Start', type_='Prior', feeders=['Retained Earnings']),
    Account('Net Income'),
    Account('Dividends Paid'),
    Account('Retained Earnings', type_='Sum', feeders=['Net Income', 'Dividends Paid', 'Retained Earnings - Start']),
    Account('Equity Capital - Start', type_='Prior', feeders=['Equity Capital']),
    StockSaleProceeds(),
    Account('Equity Capital', type_='Sum', feeders=['Stock Sale Proceeds', 'Equity Capital - Start' ]),
    Account("Shareholders' Equity", type_='Sum', feeders=['Retained Earnings', 'Equity Capital']),
])

CFLOW_ACCOUNTS = CFlowAccounts([
    Account('Cash - Start', type_='Prior', feeders=['Cash - End']),
    Account('Net Income'),
    Account('Amortization'),
    DeltaWoCap(),
    Account('Cash from Operations', type_='Sum', feeders=['Net Income', 'Amortization', DeltaWoCap.FULL_NAME]),
    PurchasePPE(),
    Account('Proceeds from Sale'),
    Account('Cash from Investing', type_='Sum', feeders=[PurchasePPE.FULL_NAME, 'Proceeds from Sale']),
    Account('Proceeds of LTD'),
    Account('Repayment of LTD'),
    Account('Dividends Paid'),
    StockSaleProceeds(),
    Account('Cash from Financing', type_='Sum', feeders=['Proceeds of LTD', 'Repayment of LTD', StockSaleProceeds.FULL_NAME]),
    Account('Change in Cash', type_='Sum', feeders=['Cash from Operations', 'Cash from Investing', 'Cash from Financing']),
    Account('Shadow Cash', hide=True),
    Account('Cash - End', type_='Equal', feeders=['Shadow Cash']),  
])



import numpy as np
import pandas as pd
from collections import UserString

class ColorStr(UserString):
    @property
    def nohash(self):
        return self.data[1:]
    
    def hex_to_rgb(self): 
        return tuple(int(self.nohash[i:i+2], 16) for i in (0, 2, 4))

    def rgb_to_bgrhex(self, rgb):
        '''
        win32 uses bgr in hex
        '''
        bgr = (rgb[2], rgb[1], rgb[0])
        strValue = '%02x%02x%02x' % bgr
        iValue = int(strValue, 16)

        return iValue

    @property
    def bgrhex(self):
        if not hasattr(self, '_bgrhex'):
            self._bgrhex = self.rgb_to_bgrhex(self.hex_to_rgb())
        
        return self._bgrhex

class grc:
    blue = ColorStr('#253765')
    green = ColorStr('#6dbc45')
    lblue = ColorStr('#2765b0')
    drkgreen = ColorStr('#335B20')
    vlblue = ColorStr('#c5d9f3')
    lgreen = ColorStr('#95CF77')

    # vlblue = ColorStr('#eef4fb')

    @classmethod
    def colors(cls, nohash=False):
        colors = [cls.blue, cls.green, cls.lblue, cls.drkgreen, cls.vlblue, cls.lgreen]

        if nohash:
            return [c.nohash for c in colors]
        else:
            return colors

class tabs:
    @classmethod
    @property
    def inputs(cls):
        return np.array([        
            'Assumptions',
            'Pipeline',
            'Headcount - Plan',
            'Headcount - Existing',
            'Opex',
            'Capex',
            'Capital'
        ])
    
    @classmethod
    @property
    def spacers(cls):
        return np.array(['Inputs --->', 'Forecast --->', 'Annual --->', 'Quarterly --->', 'Supporting --->'])
    
    @classmethod
    @property
    def order(cls):
        return np.concatenate((cls.inputs, cls.spacers))
    
    @classmethod
    @property
    def before_after(cls):
        return pd.concat((
            pd.Series(cls.order), 
            pd.Series(cls.order).shift(-1), 
            pd.Series(cls.order).shift(1)
            ), 
            axis=1
        ).set_index(0).rename(columns={1: 'before', 2: 'after'})
    
    @classmethod
    def set_color(cls, sheet):
        if sheet.name in tabs.spacers:
            color = grc.blue
        elif sheet.name in tabs.inputs:
            color = grc.green
        else:
            color = grc.vlblue
        
        sheet.api.Tab.Color = color.bgrhex



import datetime as dt
import numpy as np
import pandas as pd

import xlwings as xw

WAGE_INFLATION = .03
IDX_COL_NAMES = ['Role', 'Segment', 'Focus', 'Salary']



def wagemaker(plan):
    headix, = np.where((plan.iloc[:, :4] == IDX_COL_NAMES).all(axis=1))
    splits = np.split(plan, headix)
    sgmts = []
    for df_sgmt in splits[1:]:
        df_sgmt.columns = df_sgmt.iloc[0]

        df_sgmt = df_sgmt.iloc[1:]
        footidx, = np.where(df_sgmt.Role == 'Total')

        df_sgmt = df_sgmt.iloc[:footidx[0]]
        df_sgmt = df_sgmt[df_sgmt.columns[~df_sgmt.columns.astype('str').str.contains('202')]]

        df_sgmt.Focus = df_sgmt.Focus.fillna('')

        front = np.arange(4)
        back = np.arange(df_sgmt.shape[1] - 16, df_sgmt.shape[1])
        other = np.array([df_sgmt.columns.get_loc('Current')])

        icols = np.concatenate((front, other, back))
        df_sgmt = df_sgmt.iloc[:, icols].set_index(IDX_COL_NAMES).fillna(0)
        df_sgmt.columns.name = 'Period'
        sgmts.append(df_sgmt)

    df = pd.concat(sgmts)

    df_starts = df.iloc[:, 1:].applymap(type_sched)
    expands = []
    for i in range(df_starts.shape[1]):
        df_exp = df_starts.iloc[:, i].apply(pd.Series)
        expands.append(df_exp)

    df_starts = pd.concat(expands, axis=1).astype('int')
    df_starts.columns = pd.period_range(start='2022-01', periods=df_starts.shape[1], freq='M')
    df = pd.concat((df.iloc[:, :1], df_starts), axis=1).astype('int')

    factors = _wage_factors(*df.shape)
    factors = np.hstack((factors[:, :1], factors))
    wo_inf = _wages_wo_inf(df.columns, df.index.get_level_values('Salary').astype('float')/12)
    wages_w_inf = np.multiply(wo_inf.astype('float'), factors)
    df_wages = np.multiply(wages_w_inf, df.cumsum(axis=1))

    return df, df_wages

def make_raw(plan, type_=''):
    df, df_wages = wagemaker(plan)

    if type_ == 'wages':
        df_wages = df_wages.rename(columns={'Current': '2021-12'})
        df_wages.columns = pd.PeriodIndex(df_wages.columns)
        df_wages.columns = df_wages.columns.strftime('%b-%y')

        return df_wages
    elif type_ == 'hc':
        df = df.rename(columns={'Current': '2021-12'})
        df.columns = pd.PeriodIndex(df.columns)
        df.columns = df.columns.strftime('%b-%y')
        
        return df

def resamp(df, type_='', freq=''):
    rsamp_str = '%b-%y' if freq == 'Q' else '%Y'
    aggfunc = 'sum' if type_ == 'wages' else 'last'
    df.columns = pd.PeriodIndex([dt.datetime.strptime(s, '%b-%y') for s in df.columns], freq='M')
    resampled = df.T.resample(freq).agg(aggfunc).T.groupby('Segment').sum()
    resampled.loc['Total'] = resampled.sum()
    resampled.columns = resampled.columns.strftime(rsamp_str)
    return resampled

@xw.func
def quarterly_rev():
    wb = xw.Book.caller()
    pipeline = wb.sheets['Pipeline']
    pipeline = pipeline['A1:AU120'].options(pd.DataFrame, index=False).value
    rev = RevStat.from_schedule(pipeline)
    rev = rev._base.isutils.accts().loc['Revenue']

    return rev.values

@xw.func
def annual_rev():
    wb = xw.Book.caller()
    pipeline = wb.sheets['Pipeline']
    pipeline = pipeline['A1:AU120'].options(pd.DataFrame, index=False).value
    rev = RevStat.from_schedule(pipeline)
    rev = rev._base.isutils.resample('A', accts=True).loc['Revenue']

    return rev.values

@xw.func
@xw.arg('plan', pd.DataFrame, index=False, header=False, ndim=2)
@xw.ret(expand='table')
def raw_wages(plan):
    return make_raw(plan, 'wages')

@xw.func
@xw.arg('plan', pd.DataFrame, index=False, header=False, ndim=2)
@xw.ret(expand='table')
def raw_headcount(plan):
    hc = make_raw(plan, 'hc')
    return hc.cumsum(axis=1)
    
@xw.func
@xw.arg('wages', pd.DataFrame, index=False, header=True, expand='table', ndim=2)
def annual_wages(wages):
    ### Split, sort, and recombine
    wages = wages.set_index(IDX_COL_NAMES)
    return resamp(wages, 'wages', 'A').iloc[:, 1:]

@xw.func
@xw.arg('wages', pd.DataFrame, index=False, header=True, expand='table', ndim=2)
def quarterly_wages(wages):
    ### Split, sort, and recombine
    wages = wages.set_index(IDX_COL_NAMES)
    return resamp(wages, 'wages', 'Q').iloc[:, 1:]

@xw.func
@xw.arg('hc', pd.DataFrame, index=False, header=True, expand='table', ndim=2)
def annual_hc(hc):
    hc = hc.set_index(IDX_COL_NAMES)
    return resamp(hc, 'hc', 'A')

@xw.func
@xw.arg('hc', pd.DataFrame, index=False, header=True, expand='table', ndim=2)
def quarterly_hc(hc):
    hc = hc.set_index(IDX_COL_NAMES)
    return resamp(hc, 'hc', 'Q')

@xw.func
@xw.arg('existing', pd.DataFrame, index=True, header=True, ndim=2)
def existing_counts(existing):
    existing.Focus = existing.Focus.fillna('')
    counts = existing.groupby(['Segment', 'Role', 'Focus', 'Salary']).size()
    counts.columns = ['Count']
    return counts

@xw.func
@xw.arg('counts', pd.DataFrame, index=False, header=True, ndim=2)
def existing_lookup(sgmt, role, focus, salary, counts):
    focus = '' if focus is None else focus
    counts = counts.set_index(['Segment', 'Role', 'Focus', 'Salary'])
    if counts.index.isin([(sgmt, role, focus, salary)]).any():
        return counts.loc[(sgmt, role, focus, salary)].iloc[0]
    else:
        return 0

@xw.func
@xw.arg('hires', pd.Series, index=False, header=False, ndim=1)
def find_schedule(hires):
    hires = pd.DataFrame(hires.fillna(0).values.reshape(-1, 4)).T
    cumsched = hires.apply(sum_sched_for_series).values.cumsum()
    if (cumsched == 0).all():
        return None
    else:
        return ', '.join([f'{v:.0f}' for v in cumsched])

def type_sched(x):
    if isinstance(x, str):
        splitter = x.split(', ') if ', ' in x else x.split(' ')
        x = np.array([int(h) for h in x.split(' ') if h])
        if x.size > 3:
            x[2] = x[2:].sum()
            x = x[:3]
        elif x.size < 3:
            x = np.concatenate((x, np.zeros(x.size - 1)))
    elif isinstance(x, (int, float)):
        x = np.array([x, 0 , 0])
    else:
        raise

    return x

def sum_sched(x):
    if isinstance(x, str):
        x = sum([int(h) for h in x.split(' ') if h])
    elif isinstance(x, (int, float)):
        pass
    else:
        raise

    return x

def sum_sched_for_series(ser):
    ser = ser.apply(sum_sched)
    return ser.sum()

def _wage_factors(hcsize, projsize):
    wage_factors = np.array([(1 + WAGE_INFLATION)**i for i in np.arange(projsize//12)])
    wage_factors = np.repeat(wage_factors, 12)

    return np.tile(wage_factors, (hcsize,1))

def _wages_wo_inf(projrng, wages):
    return np.repeat(wages, projrng.size).values.reshape((wages.size, projrng.size))

@xw.func
@xw.arg('recruits', pd.DataFrame, index=False, header=True, ndim=2, expand='table')
@xw.ret(expand='table')
def recruit_fees(recruits):
    # plan = wb.sheets['Headcount - Plan'].range('A10').expand().options(pd.DataFrame, index=False).value

    recruits = recruits.set_index(recruits.columns[:9].tolist()).fillna(0)
    recruits = recruits[recruits.columns[recruits.columns.str.contains('Q')]]
    fees = recruits.index.get_level_values('Recruiting Fee').fillna(0).values.astype('float').reshape(-1,1)
    recruits.loc[:, :] = fees * recruits.values
    return recruits.sum(axis=0).values


import numpy as np
import functools as ft

from numpy.lib.utils import deprecate_with_doc

def testit(func):
    @ft.wraps(func)
    def wrapper(self, *args, **kwargs):
        try:
            left, right = func(self, *args, **kwargs)
            assertion = np.isclose(left, right, atol=1e-3) 
            assert assertion.all()
            self._results[func.__name__] = True
        except AssertionError as e:
            self._results[func.__name__] = False
            self._fails.append(FailContainer(func.__name__, assertion, left, right))
        except Exception as e:
            print (func.__name__)
            raise e

    return wrapper

class BaseTests:
    def __init__(self):
        self._results = {}
        self._fails = []

    @property
    def results(self):
        return self._results
    
    @property
    def fails(self):
        return self._fails
    
    # MAIN TESTING METHODS #
    @property
    def tests(self):
        return [f for f in self.__dir__() if 'test_' in f]
    
    def run(self):
        resprint = ''
        for i in range(len(self.tests)):
            test = self.tests[i]
            testfunc = getattr(self, test)
            testfunc()
            
            if self.results[testfunc.__name__]:
                resprint += '.'
            else:
                resprint += 'F'

            if i == len(self.tests) - 1:
                print (resprint)
            else:
                print (resprint, end='\r')
        
        n_failed = len(self.results) - np.array(list(self.results.values())).sum()
        n_passed = len(self.results) - n_failed

        print (f'Passed: {n_passed}  Failed: {n_failed}')

        if not all(self.results.values()):
            print ('Follwing tests were failed: ' + ' '.join([f.__repr__() for f in self.fails]))

class FailContainer:
    def __init__(self, func_name, boolarr, left, right):
        self.func_name = func_name
        self.boolarr = boolarr
        self.left = left
        self.right = right
        self.diff = left - right
        self.results = [func_name, boolarr, left, right]
    
    def __repr__(self):
        return self.func_name

class IStatTests(BaseTests):
    name = 'IStatTests'
    def __init__(self, istat):
        self.istat = istat
        super().__init__()
    
    # Test EBIT #
    @property
    def ebitda(self):
        return self.istat.stat.loc[('EBITDA', '---'), :].values 

    @property
    def amort(self):
        return self.istat.stat.loc['Amortization', :].sum().values 

    @property
    def ebit(self):
        return self.istat.stat.loc[('EBIT', '---'), :].values 

    @testit
    def test_ebit(self):
        return self.ebitda - self.amort, self.ebit

    @property
    def ebt(self):
        return self.istat.stat.loc[('EBT', '---'), :].values

    @property
    def interest(self):
        return self.istat.stat.loc[('Interest', '---'), :].values

    @testit
    def test_ebt(self):
        return self.ebit - self.interest, self.ebt

    @property
    def taxes(self):
        return self.istat.stat.loc[('Taxes', '---'), :].values

    @property
    def net_income(self):
        return self.istat.stat.loc[('Net Income', '---'), :].values

    @testit
    def test_ni(self):
        return self.ebt - self.taxes, self.net_income

class BSheetTests(BaseTests):
    name = 'BSheetTests'
    def __init__(self, bsheet):
        self.bsheet = bsheet
        
        super().__init__()
    
    def roller(self, arr, rolls:int):
        arrs = np.tile(arr, (rolls, 1))

        for i in np.arange(arrs.shape[0]):
            if i > 0:
                arrs[i] = np.roll(arrs[i], i)
                arrs[i, :i] = 0

        return arrs
    
    # TEST FOR AR #
    @property
    def ar_new(self):
        return self.bsheet._isprox.loc['Total Revenue'].values
    
    @property
    def ar(self):
        return self.bsheet.stat.loc['Accounts Receivable', :].values
    
    @property
    def ar_n(self):
        return self.bsheet.AR_COLLECTION.n
    
    @testit
    def test_ar(self):
        ars = self.roller(self.ar_new, self.ar_n)
        return ars.sum(axis=0), self.ar

    # TEST FOR INV #
    @property
    def inv_n(self):
        return self.bsheet.INVENTORY_LEAD.n
    
    @property
    def cogs_for_inv(self):
        return self.bsheet._isprox.loc['COGS'].shift(-self.inv_n).fillna(0)
    
    @property
    def inv(self):
        return self.bsheet.stat.loc['Inventory', :].values[0]
    
    @testit
    def test_inv(self):
        invs = self.roller(self.cogs_for_inv, self.inv_n)
        return invs.sum(axis=0)[:-2], self.inv[:-2]
    
    # Test for AP #
    @property
    def ap_n(self):
        return self.bsheet.AP_COLLECTION.n

    @property
    def ap(self):
        return self.bsheet.stat.loc['Accounts Payable', :].values[0]
    
    @testit
    def test_ap(self):
        aps = self.roller(self.cogs_for_inv, self.ap_n)
        return aps.sum(axis=0)[:-2], self.ap[:-2]

    # Test Cash #
    @property
    def shadow_cash(self):
        return self.bsheet.stat.loc['Shadow Cash', :].values[0]

    @property
    def cash(self):
        return self.bsheet.stat.loc['Cash', :].values[0]

    @property
    def indebt(self):
        return self.bsheet.stat.loc['Indebtedness', :].values[0]

    @testit
    def test_cash(self):
        cash_diff = self.shadow_cash - self.cash + self.indebt
        return cash_diff, np.zeros(cash_diff.shape[0])

    # Test Current Assets #
    @property
    def ca_from_wocass(self):
        return self.bsheet.stat.loc[self.bsheet.WOCASS[:-1].names].sum().values
    
    @property
    def ca(self):
        return self.bsheet.stat.loc['Current Assets'].values[0]

    @testit
    def test_ca(self):
        return self.ca_from_wocass, self.ca

    # Test Current Liabilities #
    @property
    def cl_from_woclia(self):
        return self.bsheet.stat.loc[self.bsheet.WOCLIA.names].sum().values

    @property
    def cl(self):
        return self.bsheet.stat.loc['Current Liabilities'].values[0]

    @testit
    def test_ca(self):
        return self.cl_from_woclia, self.cl

    # Test Balance #
    @property
    def bal(self):
        return self.bsheet.stat.loc['bal', :].values[0]
    
    @testit
    def test_bal(self):
        return np.round(self.bal[:-2], -2), np.zeros(self.bsheet.stat.columns[:-2].size)

    # Test Taxes #
    @property
    def ebt(self):
        return self.bsheet._isprox.loc['EBT']
        
    @testit
    def test_taxes(self):
        return self.ebt.values*self.bsheet.TAX_RATE, self.bsheet.stat.loc['Taxes'].values

    # Test TaxFwd #
    @property
    def cumtax(self):
        return self.bsheet.stat.loc['Taxes'].cumsum(axis=1).values[0]

    @property
    def taxfwd(self):
        return self.bsheet.stat.loc['Tax Carry-Forward'].values[0]

    @testit
    def test_taxfwd(self):
        black_tax_mask = self.cumtax > 0
        return self.taxfwd[black_tax_mask], 0

    # Test Amortization and Capital Assets #
    @property
    def capass(self):
        return self.bsheet.stat.loc['Capital Assets'].values[0]

    @property
    def ppe(self):
        return self.bsheet.stat.loc['Purchases of PPE'].values[0]

    @property
    def amort(self):
        return self.bsheet.istat._amort.stat.sum(axis=0).values

    @testit
    def test_amort(self):
        return self.capass[self.bsheet._ledge:], self.ppe[self.bsheet._ledge:].cumsum() - self.amort.cumsum()[:self.ppe[self.bsheet._ledge:].size]

class CFlowTests(BaseTests):
    name = 'CFlowTests'
    def __init__(self, cflow):
        self.cflow = cflow
        
        super().__init__()

    # Test Start / End Cash #
    @property
    def cash_start(self):
        return self.cflow.stat.loc['Cash - Start'].values[0]
    
    @property
    def cash_end(self):
        return self.cflow.stat.loc['Cash - End'].values[0]

    @property
    def delta_cash(self):
        return self.cflow.stat.loc['Change in Cash'].values[0]

    @testit
    def test_cash_ends(self):
        return self.cash_start[:-2] + self.delta_cash[:-2], self.cash_end[:-2]

class FStatTests:
    def __init__(self, istat=None, bsheet=None, eqstat=None, cflow=None):
        self.istat = IStatTests(istat) if istat is not None else istat
        self.bsheet = BSheetTests(bsheet) if bsheet is not None else bsheet
        # self.eqstat = EQStatTestseqstat
        self.cflow = CFlowTests(cflow) if cflow is not None else cflow
    
    @property
    def tests(self):
        return [self.istat, self.bsheet, self.cflow] # self.eqstat,

    def run(self):
        for t in self.tests:
            print (t.name)
            t.run()
            print ('')



import numpy as np
import pandas as pd
from typing import Iterable

import scipy.optimize as sciopt
import scipy.stats as scist

def cagr(fv, pv, t):
    """
    Computes compound annual growth rate implied by two values over a time, t

    Params
    -------
    fv:     float, future value
    pv:     float, present value
    t:      time between values in years
    """
    return (fv / pv)**(1/t) - 1

def npv(cf, disc=.08):
    """
    Net Present Value calculation for a stream of cash flows
    
    Params
    -------
    cf:    nxm array of cash flow streams
            > n = # of years
            > m = # of different cash flow streams
    disc:  float representing annual discount rate
            > default value per Duff&Phelps recommendation here:
                https://www.duffandphelps.com/-/media/assets/pdfs/publications/articles/dp-erp-rf-table-2020.pd
    
    Returns
    --------
        nxm array representing of each cash inflow discounted to the current day
    """
    return cf.values.reshape(-1,cf.ndim) / (1+disc)**np.arange(0, cf.shape[0]).reshape(-1,1)

def logparams(x1, y1, x2, y2):
    """
    Returns parameters for line of form 
    
    y = a*ln(b*x)
    
    https://math.stackexchange.com/questions/716152/graphing-given-two-points-on-a-graph-find-the-loglarithmic-function-that-passes
    """
    a = (y1 - y2) / np.log(x1 / x2)
    b = np.exp((y2*np.log(x1) - y1*np.log(x2)) / (y1 - y2))
    
    return a, b

def logline(x, a, b):
    xdim = b.shape[0]
    x = x.repeat(xdim).reshape(-1,xdim)
    a = a.reshape(-1,1)
    b = b.reshape(-1,1)

    return a*np.log(b*x.T)

def log_returns(ser):
    if not isinstance(ser, pd.Series):
        ser = pd.Series(ser)
    return np.log(ser / ser.shift(1)).iloc[1:]

def exponparams(x1,y1,x2,y2):
    """    
    $
    y_1 = b*a^{x_1}
    \\y_2 = b*a^{x_2}
    \\
    \\b = \frac{y_1}{a^{x_1}}
    \\
    \\y_2 = \frac{y_1}{a^{x_1}}a^{x_2}
    \\a^{x_2}/a^{x_1} = y_2/y_1
    \\a^{x_2 - x_1} = \frac{y_2}{y_1}
    \\a = (\frac{y_2}{y_1})^\frac{1}{x_2-x_1}
    $

    $
    \\b = \frac{y_1}{((\frac{y_2}{y_1})^\frac{1}{x_2-x_1})^{x_1}}
    \\b = \frac{y_1}{(\frac{y_2}{y_1})^\frac{x_1}{x_2-x_1}}
    $"""
    return calc_a(x1,y1,x2,y2), calc_b(x1,y1,x2,y2)

def calc_b(x1,y1,x2,y2):
    num = y1
    power = x1/(x2-x1)
    denom = (y2/y1)**power
    return num / denom

def calc_a(x1,y1,x2,y2):
    return (y2/y1)**(1/(x2-x1))

def exponline(x, a, b, c):
    if isinstance(b, np.ndarray):
        xdim = b.shape[0]
        x = x.repeat(xdim).reshape(-1,xdim)

    return b*(a**x) + c

def expon(x, a, b, c):
    return b * np.exp(a * x) + c

def expon_fit(y, x:Iterable=None):
    x = x if x is not None else np.arange(1, y.shape[0] + 1)
    params, covmat = sciopt.curve_fit(expon, x, y, bounds=(0, np.inf))

    par1_str = f'{params[1]:,.1f}' if params[1] < 1000 else \
        f'{params[1]:.1e}'
    eq = f'{par1_str}$e^{{{params[0]:,.2f}x}}$'

    y_ = expon(x, *params)
    resids = y - y_
    rmse = np.sqrt(np.mean(resids**2))
    resid_fit = scist.norm.fit(resids)
    
    return params, y_, resids, rmse, resid_fit, eq

def arth2geo(arth, std):
    """
    Accepted approximation of the geometric mean given the arthmetic mean and 
    the standard deviation of a series.
    """
    return arth - (.5*std**2)

### TRIGONOMETRY ###
def sech(x, k):
    return 1 / np.cosh(k*x)

def csch(x):
    return 1 / np.sinh(x)

def coth(x):
    return np.cosh(x) / np.sinh(x)

def cosh_squared(x,k):
    return .5*(np.cosh(2*k*x) + 1)

def sech_squared(x,k):
    return 1 / cosh_squared(x, k)

def bendit(x, x1, y1, x2, y2, m=1):
    num = m**(x - x1) - 1
    denom = m**(x2 - x1) - 1
    return (y2-y1)*(num/denom) + y1

def sig(x, a=0, b=1, c=0, d=1):
    return a + (b / (1 + np.exp(-(x-c)/d)))

def revsig(x, a=0, b=1, c=0, d=1):
    # insnum = a - x
    # insdenom = x*b
    # num = - np.log(insnum/insdenom)
    # denom = np.log(c)
    return c - d*(np.log(b / (x-a)) - 1)
    # return num/denom




"""
Pandas Accessors with Helper Functions for 
Aggregating and Manipulating Statements
"""
import functools as ft
import pandas as pd

class UtilsMixin:
    ### FStat Extension Utilities for Statements; 
    ### This is mixed in to Statement Objects (NOT a pandas accessor) ###
    def __init__(self, *args, **kwargs):
        from finstat.finstat import BaseBSheet, BaseIStat

        if isinstance(self, BaseIStat):
            self.UTILITY = IStatUtils
        elif isinstance(self, BaseBSheet):
            self.UTILITY = BSheetUtils
        else:
            raise ValueError(f'Statement not recognized {type(self)}')
        
    @property
    def utils(self):
        if not hasattr(self, '_utils'):
            self._utils = self.UTILITY(self)
        
        return self._utils

    def resample(self, *args, **kwargs):
        if hasattr(self, 'aggfuncs'):
            kwargs['aggfunc'] = self.aggfuncs

        return self.utils.resample(stat=self.stat, *args, **kwargs)

    def sgmts(self):
        return self.utils.sgmts()

    def accts(self):
        return self.utils.accts()

    def annually(self, *args, **kwargs):
        if hasattr(self, 'aggfuncs'):
            kwargs['aggfunc'] = self.aggfuncs
        return self.utils.annually(*args, **kwargs)

    def quarterly(self, *args, **kwargs):
        if hasattr(self, 'aggfuncs'):
            kwargs['aggfunc'] = self.aggfuncs
        return self.utils.quarterly(*args, **kwargs)

class FStatUtils:
    """
    Pandas extension to add common styles to dataframes in Jupyter Notebook
    """
    RANK = {
        'A-DEC': 2,
        'Q-DEC': 1,
        'M': 0,
    }
    DIVISOR = {
        'A-DEC': {'Q-DEC': 4, 'M': 12},
        'Q-DEC': {'M': 3}
    }

    def __init__(self, obj):
        self._statobj = obj
        self._obj = obj.stat

    def parse_period(self, period):
        if period in ['A', 'a', 'annual', 'annually']:
            period = 'A-DEC'
        if period in ['Q', 'q', 'quarter', 'quarterly']:
            period = 'Q-DEC'
        if period in ['M', 'm', 'month', 'mon', 'monthly']:
            period = 'M'
        
        return period

    def by_acct(self, acct, drop=False):
        if self._obj.index.nlevels > 1:
            return self._obj.xs(acct, level='Account', drop_level=drop)
        else:
            return self._obj.loc[acct]

    def annually(self, period='A-DEC', **kwargs):
        return self.resample(period=period, **kwargs)

    def quarterly(self, period='Q-DEC', **kwargs):
        return self.resample(period=period, **kwargs)

    def monthly(self, period='M', **kwargs):
        return self.resample(period=period, **kwargs)

    def sgmts(self):
        if 'Segment' not in self._obj.index.names:
            raise ValueError('This statement does not have Segments')

        return self._obj.groupby(['Segment', 'Account']).sum()

    def accts(self):
        return self._obj.groupby('Account').sum()

    def clip_rng(self, df, clip_rng):
        if clip_rng and isinstance(clip_rng, pd.PeriodIndex):
            colmask = clip_rng
        else:
            colmask = (self._statobj._START <= df.columns) & (self._statobj._END >= df.columns)
        return df.loc[:, colmask]

def resampler(func):
    @ft.wraps(func)
    def wrapper(self, period, sgmts=False, accts=False, clip_rng=[], str_cols=False, stat=None, **kwargs):
        if stat is not None:
            self._obj = stat

        if sgmts:
            self._temp = self.sgmts()
        elif accts:
            self._temp = self.accts()
        else:
            self._temp = self._obj

        if clip_rng:
            self._temp = self.clip_rng(self._temp, clip_rng)

        self._temp = self._temp.sort_index()

        period = self.parse_period(period)
        self._temp = func(self, period, **kwargs)
        if str_cols:
            strf = '%Y' if period == 'A-DEC' else '%b-%y'
            self._temp.columns = self._temp.columns.strftime(strf)

        return self._temp

    return wrapper

class IStatUtils(FStatUtils):
    @resampler
    def resample(self, period, aggfunc='', **kwargs):
        if aggfunc:
            return self._temp.T.resample(period).agg(aggfunc).T
        else:
            if self.RANK[period] >= self.RANK[self._temp.columns.freqstr]:
                return self._temp.T.resample(period).sum().T
            else:
                divint = self.DIVISOR[self._temp.columns.freqstr][period]
                return (self._temp.T.resample(period).asfreq().T / divint).ffill(axis=1)

class BSheetUtils(FStatUtils):
    @resampler
    def resample(self, period, aggfunc='', **kwargs):
        if self.RANK[period] >= self.RANK[self._temp.columns.freqstr]:
            if isinstance(aggfunc, list):
                stats = []
                for func in aggfunc:
                    stat = self._temp.loc[list(func.values())[0]].T.resample(period).agg(list(func.keys())[0]).T
                    stats.append(stat)
                df = pd.concat(stats)
                newidx = df.index.to_frame().reset_index(drop=True)
                newidx.Account = pd.CategoricalIndex(newidx.Account, categories=self._temp.index.get_level_values(0).categories, ordered=True)
                multindex = pd.MultiIndex.from_frame(newidx)
                df.index = multindex
                self._temp = df.sort_index()
                return self._temp
            else:
                aggfunc = 'last' if not aggfunc else aggfunc
                return self._temp.T.resample(period).agg(aggfunc).T
        else:
            raise ValueError('Balance Sheet cannot be downsampled')

@pd.api.extensions.register_dataframe_accessor('isutils')
class IStatExt(IStatUtils):
    """
    Pandas extension to add common styles to dataframes in Jupyter Notebook
    """
    def __init__(self, obj):
        self._obj = obj

@pd.api.extensions.register_dataframe_accessor('bsutils')
class BSheetExt(BSheetUtils):
    """
    Pandas extension to add common styles to dataframes in Jupyter Notebook
    """
    def __init__(self, obj):
        self._obj = obj



import numpy as np
import pandas as pd
import functools as ft

from finstat.accounts import ISTAT_ACCOUNTS, BSHEET_ACCOUNTS, EQ_ACCOUNTS, CFLOW_ACCOUNTS
from finstat.fincalcs import expon, expon_fit
from finstat.extensions import UtilsMixin, IStatUtils
from finstat.visuals import Plots

# UTILITIES, ASSIGNORS, AND METRIC FUNCS
def set_assumps(func):
    # Sets constants for object if 'assumps' key passed
    @ft.wraps(func)
    def wrapper(self, *args, **kwargs):
        if 'assumps' in kwargs and isinstance(kwargs['assumps'], pd.DataFrame):
            assumps = self.constants_as_records(kwargs['assumps'])
            for name, value in assumps.items():
                if name in self.INT_ASSUMPS:
                    value = int(value['Factor'])
                elif name in self.FREQ_ASSUMPS:
                    value = int(value['Factor'])
                    value = pd.tseries.frequencies.to_offset(f'{value}M')
                else:
                    value = value['Factor']

                setattr(self, name, value)
            
        return func(self, *args, **kwargs)
    
    return wrapper

def set_projrng(func):
    @ft.wraps(func)
    def wrapper(self, *args, **kwargs):
        if 'projrng' in kwargs and kwargs['projrng'] is not None:
            self._projrng = kwargs['projrng']
        
        return func(self, *args, **kwargs)
    
    return wrapper

### INDEX MIXIN ###
class IndexMix:
    """Mixin for index sublcass of each statement"""
    def __init__(self, obj):
        self.obj = obj
        self._levels = self._set_levels()
        self._base = self._make_base()

    @property
    def nlevels(self):
        return len(self.levels)
    
    @property
    def acctpos(self):
        return self.names.index('Account')
    
    @property
    def levels(self):
        return self._levels
    
    @property
    def names(self):
        return list(self.levels.keys())
    
    @property
    def base(self):
        return self._base

    def _set_levels(self):
        levels = {k:v for k, v in self.__class__.__dict__.items() if '__' not in k}
        if 'Account' not  in levels:
            levels['Account'] = ISTAT_ACCOUNTS.names
        return levels
    
    def _make_base(self):
        idxs = {}
        for name, cats in self.levels.items():
            if name == 'Segment' and self.obj.is_istat:
                cats = self.obj.sgmt_categories()

            any_cats = cats.size if isinstance(cats, np.ndarray) else cats
            if any_cats:
                idxs[name] = pd.Categorical([], cats, ordered=True)
            else:
                idxs[name] = cats

        codes = [[] for i in range(len(idxs))]

        return pd.MultiIndex(levels=idxs.values(), codes=codes, names=idxs.keys())
 
### BASE STATEMENTS ###
class BaseStat(UtilsMixin):
    _START = '2022-01'
    _END = '2025-12'
    _projrng = pd.period_range(start=_START, end=_END, freq='Q')
    FREQKEY = {'H': '2Q', 'Q': 'Q', 'M': 'M', 'A': 'A'}
    ZERO_ROWS = []
    INT_ASSUMPS = ['AMORT_YEARS']
    FREQ_ASSUMPS = ['AR_COLLECTION', 'AP_COLLECTION', 'INVENTORY_LEAD']

    CONSTANTS_KEY = {
        'Interest Rate': 'INT_RATE',
        'Income Tax Rate': 'TAX_RATE',
        'AR Terms': 'AR_COLLECTION',
        'AP Terms': 'AP_COLLECTION',
        'Inventory Lead Time': 'INVENTORY_LEAD',
        'Amortization Schedule': 'AMORT_YEARS',
        'Payroll Taxes': 'PAYROLL_TAXES',
        'Wage Inflation': 'WAGE_INFLATION',
        'Benefits': 'ANNUAL_BENEFITS',
        'Payroll Fees': 'ANNUAL_PAYROLL_FEES',
        'Unemployment Fees': 'ANNUAL_UNEMPLOYMENT',
        "Workers' Compensation": 'ANNUAL_WORKERS_COMP',
        '401K Contributions': 'ANNUAL_401K'
    }

    def __repr__(self):
        if hasattr(self, 'FULL_NAME'):

            freqstr = self.projrng.freq.__str__()
            if ':' in freqstr:
                freqstr = freqstr.split(':')[0]
            
            freqstr = freqstr.strip('<').strip('>')
            start = self.projrng[0].strftime('%b-%y')
            end = self.projrng[-1].strftime('%b-%y')
            return f'{self.FULL_NAME}: {freqstr} {start} to {end}'
        else:
            return super().__repr__()

    @property
    def is_istat(self):
        return hasattr(self, '_is_istat') and self._is_istat

    @property
    def projrng(self):
        return self._projrng
        
    @property
    def index(self):
        return self._index
    
    @property
    def frame(self):
        return self._frame

    @property
    def stat(self):
        return self.statement

    @property
    def statement(self):
        return self._statement

    def update(self, *args, refresh=False):
        if refresh:
            self._statement = self._base.copy()

    def _make_frame(self, projrng):
        self._frame = pd.DataFrame(columns=projrng, index=self.index.base)
        return self._frame

    @classmethod
    def from_schedule(cls, *args, projrng=None, **kwargs):
        if projrng is not None:
            cls._projrng = projrng

    @classmethod
    def constants_as_records(cls, assumps):
        assumps.index = [cls.CONSTANTS_KEY[idx] if idx in cls.CONSTANTS_KEY else idx for idx in assumps.index]

        return assumps[~assumps.isna().all(axis=1)].to_dict('index')

    def _add_zero_row(self, acct_name, stat):
        if stat.index.nlevels == 2:
            stat.loc[(acct_name, '---'), :] = 0
        elif stat.index.nlevels == 1:
            stat.loc[acct_name, :] = 0

class BaseIStat(BaseStat):
    """
    Base class to provide unique features to 
    Income Statement based classes.

    In particular, UtilsMixin differentiates between BaseIStat and BaseBSheet
    """
    @set_projrng
    def __init__(self, *args, **kwargs):
        """
        All income statement components set `projrng` using the
        set_projrng decorator.
        The `super()` call then integrates the UtilsMixin
        """
        super().__init__(*args, **kwargs)

class BaseBSheet(BaseStat):
    """
    Base class to provide unique features to 
    Balance Sheet based classes.

    Currently, this base is used only to allow UtilsMixin to 
    support unique Balance Sheet utilities

    Each Balance Sheet statement determines `projrng` uniquely and so `set_projrng` is not needed
    """
    pass

### INCOME STATEMENT and SUPPORTING ###
class RevStat(BaseIStat):
    FULL_NAME = 'Sales Statement'
    ACCOUNTS = ISTAT_ACCOUNTS.by_stat('RevStat')
    _projrng = pd.period_range(start='2022-01', end=pd.Period('2025Q4', freq='Q'), freq='Q')
    
    class Index(IndexMix):
        Segment = ['APAC', 'EMEA', 'AM', '---']
        PipeType = []
        Name = []
        Confidence = ['High', 'Low', '---']
    
    def __init__(self, rev, projrng=None):
        super().__init__(self)
            
        self._index = self.Index(self)
        self._base = self._make_base(rev)
        self.update(refresh=True)
    
    def _make_base(self, rev):
        rev = self._reindex_booking(rev)
        base = self._make_frame(self.projrng)
        base = base.append(rev.stack())
        
        return base.apply(pd.to_numeric, downcast='float').sort_index()
    
    def _reindex_booking(self, rev):
        rev = rev.set_index(self.index.names[:-1])
        rev.columns = pd.MultiIndex.from_product((self.projrng, rev.columns.unique()), names=['Period', 'Account'])
        rev = rev[~rev.index.get_level_values(1).isna()]
        return rev
    
    def update(self, metrics=[], conf=[], refresh=False):
        super().update(refresh=refresh)
        
        for acct in self.ACCOUNTS.tabs:
            if acct.is_sub:
                self._statement = acct.append(self._statement)
        
        if conf:
            high_conf_mask = self._statement.index[self._statement.index.get_level_values('Confidence').isin(conf)]
            self._statement = self._statement.loc[high_conf_mask]

        self._statement = self._statement.fillna(0).sort_index()
        
        return self
    
    @classmethod
    def from_schedule(cls, rev, projrng=None):
        super().from_schedule(rev, projrng=projrng)

        rev.columns = rev.iloc[1]
        idx_start = rev.Name[rev.Name.str.find('  ') == 0].index[0] - 1
        idx_end = rev.Name[rev.Name == 'GAPP REVENUE PROJECTIONS'].index[0] - 1
        rev = rev.iloc[idx_start:idx_end]
        rev = rev[~rev.isna().all(axis=1).T]

        sgmt_brks = rev.index[rev.Name.str.find('  ') == -1]
        rev.insert(0, 'PipeType', '')
        for i in range(sgmt_brks.size):
            last = rev.index[-1] if i == len(sgmt_brks) - 1 else sgmt_brks[i + 1]
            rev.loc[sgmt_brks[i]+1:last, 'PipeType'] = rev.Name[sgmt_brks[i]]
        rev = rev.loc[rev.index.difference(sgmt_brks)]

        rev.loc[:, 'Segment'] = rev.loc[:, 'Segment'].fillna('---')
        rev.insert(2, 'Confidence', '---')

        rev = rev.T[~rev.isna().all()].T

        skiprevs = np.arange(12,4+5*4*2,10)
        skipcols = skiprevs + 1
        skips = np.sort(np.concatenate([skiprevs, skipcols]))
        keeps = np.setdiff1d(np.arange(rev.shape[1]), skips)

        rev = rev.iloc[:, keeps].fillna(0)
        
        return cls(rev)
    
class WageStat(BaseIStat):
    """
    Builds expense statement for salaried employees 
    """
    FULL_NAME = 'Salaries by Role'
    ACCOUNTS = ISTAT_ACCOUNTS.by_stat('WageStat')
    _projrng = pd.period_range(start='2022-01', end=pd.Period('2025-12', freq='M'), freq='M')

    PAYROLL_TAXES = .075
    WAGE_INFLATION = .03

    ANNUAL_BENEFITS = 6000
    ANNUAL_PAYROLL_FEES = 1200
    ANNUAL_UNEMPLOYMENT = 200
  
    ANNUAL_WORKERS_COMP = .0075
    ANNUAL_401K = .1
        
    class Index(IndexMix):
        Segment = ['CE', 'G&A', 'OPS', 'Product', 'S&M']
        Role = []
        Focus = []
        Salary = []
        ID = []
        
    def __init__(self, wages, headcount, projrng=None):
        super().__init__(self, projrng=projrng)
        
        self._index = self.Index(self)
        self._base = self._make_base(wages)

        hc = headcount.T.set_index(0).T.set_index(self.index.names[:-2])
        hc.columns = pd.PeriodIndex(pd.to_datetime(hc.columns, format='%b-%y'), freq='m')

        self._headcount = hc
        
        self.update(refresh=True)

    def _make_base(self, schedule):
        if hasattr(self, '_base'):
            raise ValueError('_base has already been made')

        base = self._make_frame(self.projrng)

        wages = self._make_wages(schedule)
        base = base.append(wages)
        
        benefits = self._make_benefits(base)
        base = base.append(benefits)
        
        return base.apply(pd.to_numeric, downcast='float').sort_index()

    def _make_wages(self, schedule):
        schedule.loc[:, 'Account'] = self.ACCOUNTS.names[0]
        schedule.loc[:, 'ID'] = np.arange(schedule.shape[0])

        if self._from == 'schedule':
            schedule.Start = pd.to_datetime(schedule.Start)
            wages_w_inf = np.multiply(self._wages_wo_inf(schedule.Salary/12), self._wage_factors(schedule.shape[0]))
            values = np.multiply(wages_w_inf, self._wagemask(schedule.Start))
            index = schedule.set_index(self.index.names).index
        elif self._from == 'wages':
            schedule = schedule.set_index(['Segment', 'Role', 'Focus', 'Salary', 'ID', 'Account'])
            index = schedule.index
            values = schedule.values[:, 1:]
        
        return pd.DataFrame(values, index=index, columns=self.projrng).fillna(0)

    def _make_benefits(self, base):
        wage_benefits = np.multiply(base, self._benefits_per_wage(base.shape))
        hc_benefits = np.multiply(base > 0, self._benefits_per_hc(base.shape))
        benefits = np.add(wage_benefits, hc_benefits)

        newidx = base.index.to_frame().reset_index(drop=True)
        newidx.Account = self.ACCOUNTS[1].name
        newidx = pd.MultiIndex.from_frame(newidx)

        return pd.DataFrame(benefits.values, index=newidx, columns=self.projrng).fillna(0)
    
    @classmethod
    @property
    def headcount_benefits(cls):
        return [cls.ANNUAL_BENEFITS, cls.ANNUAL_PAYROLL_FEES, cls.ANNUAL_UNEMPLOYMENT]

    @classmethod
    @property
    def wage_benefits(cls):
        return [cls.ANNUAL_WORKERS_COMP, cls.ANNUAL_401K]
    
    def _benefits_per_hc(self, shape):
        monthly_bens = sum(self.headcount_benefits) / 12
        return np.tile(monthly_bens, shape)

    def _benefits_per_wage(self, shape):
        monthly_bens = sum(self.wage_benefits) / 12
        return np.tile(monthly_bens, shape)
        
    def _wage_factors(self, hcsize):
        wage_factors = np.array([(1 + self.WAGE_INFLATION)**i for i in np.arange(self.projrng.size//12)])
        wage_factors = np.repeat(wage_factors, 12)
        
        return np.tile(wage_factors, (hcsize,1))

    def _wages_wo_inf(self, wages):
        return np.repeat(wages, self.projrng.size).values.reshape((wages.size, self.projrng.size))

    def _startmask(self, start):
        return (self.projrng >= start.to_period('M')).astype('int')
    
    def _wagemask(self, start):
        return np.hstack(start.apply(self._startmask).values).reshape((start.size, self.projrng.size))

    @classmethod
    @set_assumps
    def from_wages(cls, wages:pd.DataFrame, headcount:pd.DataFrame, projrng:pd.PeriodIndex=None, assumps=[]):
        cls._from = 'wages'

        super().from_schedule(projrng=projrng)
        return cls(wages.T.set_index(0).T, headcount)

    @classmethod
    @set_assumps
    def from_schedule(cls, wages:pd.DataFrame, headcount:pd.DataFrame, projrng:pd.PeriodIndex=None, assumps=[]):
        """
        Process
        --------
        Assign projection range
        Create the base dataframe
        
        The split the headcount into two main types of frames:
            1) single: roles where only one hire is required
            2) multi: roles where multiple hires are required
        
        Either type is handled separately
            1) multi: 
                + find first start for each role
                + distribute the new hires evenly during the unique role range
                + assign hire at each unique start date
            2) single:
                + self explanatory
                
        Benefits are added afterwards, some allocated based on each employees wage and some on headcount
        
        Parameters
        ------------
        wages:       Dataframe of excel wages
        """        
        cls._from = 'schedule'
        super().from_schedule(projrng=projrng)
        # Split wagess into single roles or multiple roles
        wages = wages.iloc[:,1:].sort_values(by='n')
        wages.Title = wages.Title.str.strip()
        currmask = wages.Start == 'Current'
        wages.loc[currmask, 'Start'] = cls._projrng[0].to_timestamp()
        
        single, multi = wages[wages.n == 1].copy(), wages[wages.n > 1].copy()
        
        ids = np.arange(1, wages.n.sum() + 1).tolist()
        subframes = []
        for i, role in multi.iterrows():
            role_start = pd.Period(role.Start, freq='M')
            rolerng = pd.period_range(start=role_start, end=cls._projrng[-1], freq='M')

            # Linearly distribute roles thru the rolerng. the end of the rolerng is capped by 
            # the number of periods per role, so there is always a space b/w the end and the last role start
            # the linear range of floats is then floored, which will result in some roles starting on that the same time
            start_idxs = np.floor(np.linspace(0, rolerng.size*(1-(1/role.n)), role.n)).astype('int')
            
            subframe = pd.DataFrame(np.tile(role, (role.n,1)), columns=role.index)
            subframe.Start = pd.to_datetime(rolerng[start_idxs].to_timestamp())
            
            subframes.append(subframe)
                        
        single_from_multi = pd.concat(subframes)
        single_from_multi.n = 1
        
        return cls(pd.concat([single, single_from_multi]), headcount)

class AmortSched(BaseIStat):
    FULL_NAME = 'Amortization'
    ACCOUNTS = ISTAT_ACCOUNTS.by_stat('AmortSched')
    _projrng = pd.period_range(start='2022-01', end=pd.Period('2025-12', freq='M'), freq='M')
    AMORT_YEARS = 5

    class Index(IndexMix):
        Segment = []
        Item = []
        Description = []
        Account = []
        Period = []
        
    def __init__(self, capex, projrng=None):
        super().__init__(self, projrng=projrng)

        self._capex = capex

        self._index = self.Index(self)
        self._base = self._make_base(capex)
        self.update(refresh=True)   
    
    @property
    def amortrng(self):
        return pd.period_range(
            start=self.projrng[0],
            periods=self.projrng.size + self.AMORT_MONTHS, 
            freq=self.projrng.freq
        ) 
    
    @classmethod
    @property
    def capexrng(cls):
        return pd.period_range(start=cls._projrng[0], end=cls._projrng[-1], freq='Q')
         
    @property
    def amortsched(self):
        return self._statement

    @property
    def capex(self):
        return self._capex
    
    @property
    def capexsched(self):
        return self._capex

    @classmethod
    @property
    def AMORT_MONTHS(cls):
        return 12 * cls.AMORT_YEARS

    @classmethod
    @property
    def AMORT_SCHED(cls):
        return 1 / cls.AMORT_MONTHS

    @property
    def SEGMENTS(self):
        return self.stat.index.get_level_values('Segment').unique()
    
    def _make_base(self, capex):
        asched = self._expand_sched(capex)
        base = self._make_frame(self.amortrng)
        base = base.append(asched)

        return base.apply(pd.to_numeric, downcast='float')*-1
    
    def _expand_sched(self, capex):
        # 11 asset type, across 48 monthly periods of projection, across 60 amortization months
        by_purchase = np.repeat(capex.values, self.AMORT_MONTHS).reshape(-1, capex.values.shape[1], self.AMORT_MONTHS)
        amort_by_asset = np.multiply(by_purchase, self.AMORT_SCHED)
        amort_by_asset = amort_by_asset.reshape(amort_by_asset.shape[0]*amort_by_asset.shape[1], -1)
        backend = np.zeros((amort_by_asset.shape[0], by_purchase.shape[1]))
        amort = np.hstack((amort_by_asset, backend))

        # https://stackoverflow.com/questions/20360675/roll-rows-of-a-matrix-independently
        roll = np.tile(np.arange(1,by_purchase.shape[1]+1), (by_purchase.shape[0],1)).flatten()
        rowidx, colidx = np.ogrid[:amort.shape[0], :amort.shape[1]]
        colidx = colidx - roll[:, np.newaxis]
        amortarr = amort[rowidx, colidx]
        
        newidx = self._reindex(capex)
        
        asched = pd.DataFrame(amortarr, index=newidx, columns=self.amortrng)
        
        return asched
    
    def _reindex(self, capex):
        idx = capex.index.to_frame().reset_index(drop=True)
        newidx = pd.DataFrame(np.repeat(idx.values.T, capex.columns.size).reshape(3,-1).T, columns=self.index.names[:-2])
        newidx['Account'] = 'Amortization'
        newidx['Period'] = np.tile(capex.columns.values, (idx.shape[0],1)).ravel()
        newidx = pd.MultiIndex.from_frame(newidx)
        
        return newidx
    
    @classmethod
    @set_assumps
    def from_schedule(cls, capex:pd.DataFrame, projrng=None, assumps=[]):
        super().from_schedule(projrng=projrng)
        
        capex = capex.T.set_index(0).T
        capexidx = ['Asset', 'Item', 'Description']
        capex[capexidx] = capex[capexidx].fillna('')
        capex = capex.set_index(capexidx)
        capex.columns = AmortSched.capexrng

        capex = capex.fillna(0)
        capex = (capex.T.resample('M').asfreq().T / 3).ffill(axis=1)

        return cls(capex)

class OpexStat(BaseIStat):
    FULL_NAME = 'Operating Expenses'
    ACCOUNTS = ISTAT_ACCOUNTS.by_stat('OpexStat')
        
    class Index(IndexMix):
        Segment = ['CE', 'G&A', 'OPS', 'Product', 'S&M']
        Item = []
        
    def __init__(self, istat, specs, gens, projrng=None):
        super().__init__(projrng=projrng)

        self._istat = istat        
        self._index = self.Index(self)
        self._base = self._make_base(specs, gens)

        self.update(refresh=True)
    
    @property
    def istat(self):
        return self._istat

    def top10(self):
        top10 = self.stat.sum(axis=1).sort_values(ascending=False).reset_index().set_index(['Segment', 'Item']).drop(columns='Account').iloc[:10]
        top10.columns = ['4Y Sum']
        return top10

    def exp_schedule(self):
        exp_sched = self.stat.reorder_levels(['Account', 'Segment', 'Item']).droplevel('Account').sort_index()

        items = exp_sched.index.get_level_values('Item').unique().values
        totals = exp_sched.groupby('Segment').sum()

        newidx = totals.index.to_frame()
        newidx['Item'] = 'Total'
        total = totals.sum().to_frame().T
        total.index = ['Total']
        total.index.name = 'Segment'
        totals = totals.append(total)
        totidx = totals.index.to_frame()
        totidx['Item'] = ['Total']*5 + ['---']
        totals.index = pd.MultiIndex.from_frame(totidx)
        exp_sched = exp_sched.append(totals)
        newidx = exp_sched.index.to_frame()
        newidx.Segment = pd.Categorical(newidx.Segment, newidx.Segment.unique(), ordered=True)
        newcats = np.concatenate((items, np.array(['Total', '---'])))
        newidx.Item = pd.Categorical(newidx.Item, newcats, ordered=True)
        newidx = pd.MultiIndex.from_frame(newidx)
        exp_sched.index = newidx

        return exp_sched.sort_index()

    def _make_base(self, specs, gens):
        sub_frames = self._make_gens(gens)
        specs = self._clean_specs(specs)
        sub_frames = pd.concat([specs, sub_frames])
        base = self._make_frame(self.projrng)
        base = base.append(sub_frames)
        
        return base.apply(pd.to_numeric, downcast='float')

    def _clean_specs(self, specs):
        specs['Account'] = 'Other Expenses'
        specs = specs.drop('Description', axis=1).set_index(self.index.names).fillna(0)
        specs = specs.astype('str').replace(' ', '0').astype('float')
        specs.columns = self._projrng
        
        return specs

    def _make_gens(self, gens):
        opex_subs = []
        for (var, freq), gf in gens.groupby(['Variable', 'Frequency']):
            gf['Account'] = 'Other Expenses'

            if freq in ['Q', 'A']:
                startstr = '2022-12' if freq == 'A' else '2022-03'
                start = pd.Period(startstr, freq='Q-DEC')
                opexrng = pd.period_range(start=start, end=self.projrng[-1], freq=freq)
            else:
                opexrng = self.projrng

            fctrs = gf.Factor.values

            if var == 'Salaries and Payroll Taxes':
                # BONUSES
                vars_ = self.istat.wages.utils.resample(freq, sgmts=True)
                varsmask = vars_.index.get_level_values('Segment').isin(gf.Segment)
                vars_ = vars_[varsmask].loc[pd.IndexSlice[:, var], :].values
            # Employees & New Hires
            elif var in ['Employee', 'New Hire']:
                if var == 'Employee':
                    vars_ = self.istat.wages._headcount.T.resample(freq).last().T.groupby('Segment').sum().iloc[:, 1:]
                elif var == 'New Hire':
                    vars_ = self.istat.wages._headcount.diff(axis=1).fillna(0).T.resample(self.projrng.freqstr).sum().T.groupby('Segment').sum().iloc[:, 1:]

                varsmask = vars_.index.get_level_values('Segment').isin(gf.Segment)
                vars_ = vars_[varsmask].values
            elif var in ['Revenue', 'Gross Profit']:
                # Revenue & Gross Profit
                vars_ = self.istat.rev.utils.resample(freq, accts=True)
                vars_ = vars_.loc[var].values

                fltmask = gf.Factor.apply(type) == float
                floats = gf.Factor[fltmask]
                floats = np.tile(floats, (4, 1)).T

                strs = gf.Factor[~fltmask]
                strs = (pd.DataFrame(strs.str.replace('%', '').str.replace(',', '').str.split(' ').tolist()).astype('float') / 100).values

                fctrs = np.zeros((gf.Factor.shape[0], strs.shape[1]))
                fctrs[fltmask] = floats
                fctrs[~fltmask] = strs

                reps = 4 if freq == 'Q' else 1
                fctrs = np.repeat(fctrs, reps).reshape(gf.Factor.shape[0], vars_.shape[0])

            fctrs = fctrs.reshape(-1,1) if vars_.ndim > 1 else fctrs
            values = np.multiply(vars_, fctrs.astype('float'))
            timemask = self.projrng.to_timestamp(self.projrng.freqstr).isin(opexrng.to_timestamp(opexrng.freqstr))
            timemask = np.tile(timemask, (gf.shape[0],1))
            inputs = np.zeros(timemask.size)
            inputs[timemask.flatten()] = values.flatten()
            inputs = inputs.reshape(timemask.shape)

            index = gf.set_index(self.index.names).index
            gframe = pd.DataFrame(inputs, index=index, columns=self.projrng)

            opex_subs.append(gframe)

        return pd.concat(opex_subs)
    
    @classmethod
    def from_schedule(cls, istat, opex:pd.DataFrame, projrng:pd.PeriodIndex=None):
        """
        Process
        --------
        Assign projection range
        Create the base dataframe
        
        The split the headcount into two main types of frames:
            1) single: roles where only one hire is required
            2) multi: roles where multiple hires are required
        
        Either type is handled separately
            1) multi: 
                + find first start for each role
                + distribute the new hires evenly during the unique role range
                + assign hire at each unique start date
            2) single:
                + self explanatory
                
        Benefits are added afterwards, some allocated based on each employees wage and some on headcount
        
        Parameters
        ------------
        headcount:       Dataframe of excel headcount
        """
        super().from_schedule(projrng=projrng)
        
        opexs = np.split(opex, opex[opex.isnull().all(1)].index)
        gen_opexs = [p for p in opexs if p.iloc[:, 5].isna().all()]
        spec_opexs = [p for p in opexs if ~p.iloc[:, 5].isna().all()]

        # Specific opex
        clean_specs = []
        for spec_opex in spec_opexs:
            if spec_opex.iloc[0].isna().all():
                spec_opex = spec_opex.iloc[1:]

            spec_opex.insert(0, column=-1, value=spec_opex.iloc[0,0])
            spec_opex.iloc[0,0] = 'Segment'
            spec_opex.iloc[0, 1] = 'Item'
            spec_opex = spec_opex.T.set_index(spec_opex.T.columns[0]).T
            clean_specs.append(spec_opex)

        specs = pd.concat(clean_specs)
        # General Opex
        clean_gens = []
        gen_opexs = gen_opexs if len(gen_opexs) == 5 else gen_opexs[:5 - len(gen_opexs)]
        for gen_opex in gen_opexs:
            if gen_opex.iloc[0].isna().all():
                gen_opex = gen_opex.iloc[1:]
            gen_opex.insert(0, column=-1, value=gen_opex.iloc[0,0])
            gen_opex.iloc[0,0] = 'Segment'
            gen_opex.iloc[0, 1] = 'Item'
            gen_opex = gen_opex.T.set_index(gen_opex.T.columns[0]).T.iloc[:, :6]
            clean_gens.append(gen_opex)

        gens = pd.concat(clean_gens)
        gens = gens.drop('Description', axis=1)
        
        return cls(istat, specs, gens)

class IStat(BaseIStat):
    """
    Instantiates *incomplete* income statement. Requires BSheet and CStat to complete
    """
    FULL_NAME = 'Income Statement'
    ACCOUNTS = ISTAT_ACCOUNTS
    ZERO_ROWS = ['Interest', 'Taxes']
    _projrng = pd.period_range(start='2022-01', end=pd.Period('2025-12', freq='Q'), freq='Q')
    _is_istat = True
    swapped_exp = False

    OPEX_RATS = ['sal_per_rev', 'oth_opex_per_rev', 'topex_per_rev', 'sal_v_topex', 'oth_v_topex']


    class Index(IndexMix):
        Account = ISTAT_ACCOUNTS.names
        Segment = []

    @set_assumps
    def __init__(self, revenue, wages, headcount, opex, capex, assumps=[], projrng=None):
        super().__init__(self, projrng=projrng)

        self._rev = RevStat.from_schedule(revenue)
        self._wages = WageStat.from_wages(wages, headcount, assumps=assumps)
        self._amort = AmortSched.from_schedule(capex, assumps=assumps)
        self._opex = OpexStat.from_schedule(self, opex)

        self._index = self.Index(self)
        self._base = self._make_base()
        self.update(refresh=True)
        
    @property
    def rev(self):
        return self._rev
    
    @property
    def wages(self):
        return self._wages
    
    @property
    def amort(self):
        return self._amort
    
    @property
    def opex(self):
        return self._opex
    
    @property
    def components(self):
        return [self.rev, self.wages, self.amort, self.opex]
    
    @property
    def exp_accts(self):
        return np.concatenate((self._wages.ACCOUNTS.names, self._opex.ACCOUNTS.names))
    
    def sgmt_categories(self):
        cats = []
        for c in self.components:
            sgmtidx = c.stat.index.get_level_values('Segment')
            if hasattr(sgmtidx, 'categories'):
                for cat in sgmtidx: 
                    if cat not in cats:
                        cats.append(cat)
            else:
                for cat in sgmtidx.unique():
                    if cat not in cats:
                        cats.append(cat)
                
        return cats
    
    def _make_base(self):
        base = pd.concat([c.resample(self.projrng.freqstr, sgmts=True) for c in self.components])
        base = base.loc[:, self.projrng]
        base.index = self._refresh_index(base)

        return base.reorder_levels([1,0]).sort_index()

    def _refresh_index(self, base):
        sgmts = pd.Categorical(base.index.get_level_values('Segment'), self.sgmt_categories(), ordered=True)
        accts = pd.Categorical(base.index.get_level_values('Account'), self.ACCOUNTS.names, ordered=True)
        index_as_frame = pd.DataFrame({'Segment': sgmts, 'Account': accts})
        
        return pd.MultiIndex.from_frame(index_as_frame)

    def update(self, metrics=['tr', 'tgp', 'ebit', 'ebt'], ops=[], refresh=False, hide=False):
        super().update(refresh=refresh)

        if 'opex_rats' in metrics:
            metrics.pop(metrics.index('opex_rats'))
            metrics += self.OPEX_RATS

        if refresh:
            if hasattr(self, '_bsheet'):
                for acct_name in self.ZERO_ROWS:
                    self._statement.loc[(acct_name, '---'), :] = self._bsheet._base.isutils \
                        .resample(self.projrng.freqstr).loc[acct_name, self._statement.columns].values[0]
            else:
                for acct_name in self.ZERO_ROWS:
                    if acct_name not in self._statement.index.get_level_values('Account'):
                        self._add_zero_row(acct_name, self._statement)
        
        for acct in self.ACCOUNTS.tabs:
            self._statement = acct.append(self._statement)

        for metric in metrics:
            metric_acct = self.ACCOUNTS.get(metric)
            self._statement = metric_acct.append(self._statement)            

        if ops:
            ops = [ops] if isinstance(ops, str) else ops
            for op in ops:
                self._statement = getattr(self, op)(self._statement)

        self._statement = self._statement.fillna(0).sort_index()

        if hide:
            show_mask = []
            for name in self.stat.index.get_level_values('Account'):
                acct = self.ACCOUNTS.get(name)
                if not hasattr(acct, 'show') or acct.show:
                    show_mask.append(True)
                else:
                    show_mask.append(False)
            self._statement = self._statement[show_mask]

        return self

    def backpass(self, bsheet):
        self._bsheet = bsheet
        self.update(refresh=True)

    # OPERATIONS
    def _merge_and_swap(self, stat:pd.DataFrame):
        stat = self._merge_benefits(stat)
        stat = self._swap_exp_levels(stat)

        return stat

    def _merge_benefits(self, stat:pd.DataFrame):
        wages = stat.loc[pd.IndexSlice[self._wages.ACCOUNTS.names, :]].groupby('Segment').sum()
        stat.loc[pd.IndexSlice[self._wages.ACCOUNTS.names[0], :], :] = wages.values
        stat = stat.drop(self._wages.ACCOUNTS.names[1], axis=0)
        
        return stat
    
    def _swap_exp_levels(self, stat):
        begidx = np.argwhere(self.ACCOUNTS.names == 'Gross Margin')[0,0]
        endidx = np.argwhere(self.ACCOUNTS.names == 'Total Operating Expenses')[0,0]

        swapidx = stat.index.swaplevel()

        swapped = swapidx.to_frame().loc[pd.IndexSlice[:, self.exp_accts], :].values

        stat_reset = stat.reset_index().set_index(stat.index.names, drop=False)

        new_sgmt_cats = np.concatenate((
            self._rev.index.levels['Segment'],
            self.exp_accts, 
            self._amort.stat.index.get_level_values('Segment').unique()
        ))
        new_acct_cats = np.concatenate((
            self.ACCOUNTS[:begidx].names, 
            self._wages.index.Segment,
            self.ACCOUNTS[endidx:].names
        ))
        stat_reset.loc[:, 'Account'] = pd.Categorical(stat_reset.Account, new_acct_cats, ordered=True)
        stat_reset.loc[:, 'Segment'] = pd.Categorical(stat_reset.Segment, new_sgmt_cats, ordered=True)
        
        stat_reset.loc[np.unique(swapped[:, 1]), stat.index.names] = swapped

        stat_reset = stat_reset.set_index(stat.index.names)
        stat_reset.columns = self.projrng

        return stat_reset.sort_index()

### BALANCE SHEET ###
class BSheet(BaseBSheet):
    FULL_NAME = 'Balance Sheet'
    ACCOUNTS = BSHEET_ACCOUNTS

    WOCASS = BSHEET_ACCOUNTS.accounts[[0,1,2,4]]
    WOCLIA = BSHEET_ACCOUNTS.accounts[[5,6]]
    WOCAP = WOCASS + WOCLIA
    NC_WOCAP = WOCASS[1:] + WOCLIA[:-1]

    TAX_RATE = .35
    INT_RATE = .05
    AR_COLLECTION = pd.tseries.frequencies.to_offset('3M')
    AP_COLLECTION = pd.tseries.frequencies.to_offset('2M')
    INVENTORY_LEAD = pd.tseries.frequencies.to_offset('2M')

    _projrng = pd.period_range(start='2022-01', end='2025-12', freq='M')

    class Index(IndexMix):
        Account = BSHEET_ACCOUNTS.names.tolist()
    
    @set_assumps
    def __init__(self, istat, capital, projrng=None, assumps=[], opening=None):
        super().__init__(projrng=projrng)
        self._projrng = self._set_projrng(projrng)
        
        self._index = self.Index(self)
        self._istat = istat
        self._capital = self._capital_from_sched(capital)
        self._isprox = self._set_istat_proxy(istat)
        self._base = self._make_base(opening)
        
        self.update(refresh=True, hide=False)

        self._istat.backpass(self)
 
    @property
    def _redge(self):
        return max(self.AR_COLLECTION.n, self.AP_COLLECTION.n, self.INVENTORY_LEAD.n)
    
    @property
    def _ledge(self):
        return self.INVENTORY_LEAD.n
    
    @property
    def projrng(self):
        return self._projrng
    
    @property
    def istat(self):
        return self._istat

    @property
    def capital(self):
        return self._capital

    def _set_projrng(self, projrng=None):
        projrng = projrng if projrng else self._projrng
        start = projrng[0].asfreq('M') - self._ledge
        end = projrng[-1] + self._redge
        
        return pd.period_range(start=start, end=end, freq='M')
    
    def _expand_rng(self, stat):
        new_stat = pd.DataFrame(0, index=stat.index, columns=self.projrng)
        new_stat.loc[stat.index, stat.columns] = stat.values
        new_stat = new_stat.fillna(0).sort_index()

        return new_stat
    
    def _set_istat_proxy(self, istat):
        isprox = istat.stat.isutils.resample(self.projrng.freqstr, accts=True).copy()
        isprox = self._expand_rng(isprox)
        self._extrap(isprox)
        # Must recalc amortization and EBITDA from original monthly ... not resampled
        isprox.loc['Amortization', :] = self.amortsched.values[:isprox.shape[1]]
        isprox.loc['EBIT', :] = isprox.loc['EBITDA'].values - isprox.loc['Amortization'].values
        
        return isprox

    def _extrap(self, isprox):
        accts = ['Total Revenue', 'COGS']
        
        xs = np.arange(0, self.projrng.size)
        for acct in accts:
            vals = isprox.loc[acct, :].values
            params, _, _, _, _, _ = expon_fit(vals)
            extrapd = expon(xs, *params)[-self._redge:]
            isprox.loc[acct, self.projrng[-self._redge:]] = extrapd

    def _capital_from_sched(self, capital):
        capital = capital.copy().fillna(0)
        capital.insert(1, 'Account', 'Stock Sale Proceeds')
        capital = capital.set_index(['Details', 'Account'])
        capital.columns = self.istat.projrng
        capital = capital.apply(pd.to_numeric, downcast='float')
        capital = capital.isutils.resample('M', aggfunc='first').fillna(0)

        return self._expand_rng(capital)

    def _append_opening(self, base, opening):
        opening.columns = self.projrng[:1]
        acctidx = pd.Categorical(opening.index, categories=base.index.get_level_values('Account').categories, ordered=True)
        opening.index = pd.MultiIndex.from_frame(pd.DataFrame({'Account': acctidx}))
        return base.append(opening)

    @property
    def capexsched(self):
        if not hasattr(self, '_capexsched'):
            self._capexsched = self._expand_rng(self.istat._amort.capex)
        
        return self._capexsched

    @property
    def amortsched(self):
        if not hasattr(self, '_amortsched'):
            self._amortsched = self._expand_rng(self.istat._amort.stat.sum().to_frame().T)
        
        return self._amortsched.sum()
    
    def _make_base(self, opening=None):
        #         """
        #         Transactions to be captured:
        #             + Sell product / generate revenue / ship inventory
        #             + collect AR
        #             + Purchase inventory in advance of sale; 1-month lead time
        #             + pay AP
        #             + Pay opex
        #             + make capital expenditures
        #             + amortize capex
        #             + pay interest
        #             + pay tax
        #             + retain earnings
        #             + raise capital (debt and equity)
        #         """
        base = self._make_frame(self.projrng)

        for acct in self.index.Account:
            base.loc[acct, :] = 0

        if opening is not None:
            for idx, value in opening.iterrows():
                if idx == 'Tax Carry-Forward':
                    base.loc[idx, :] = [value[0]] + [0]*(self.projrng.size - 1)
                else:
                    base.loc[idx, :] = [value[0]] * self.projrng.size

        # dr AR
        #   cr Rev
        # dr Cash
        #   cr AR 
        tr = self._isprox.loc['Total Revenue']
        invoiced = tr.cumsum() + base.loc['Accounts Receivable'].iloc[0]
        collected = invoiced.shift(self.AR_COLLECTION.n).fillna(0)
        base.loc['Accounts Receivable', :] = invoiced.values - collected.values

        # dr Inv
        #   cr AP
        # cr COGS
        #   dr Inv
        # dr AP
        #   cr cash
        # Inventory built IN ADVANCE of COGS / AP recorded same time, so must shift BACKWARDS in time
        cogs = self._isprox.loc['COGS', :]
        inv_up = cogs.cumsum().shift(-self.INVENTORY_LEAD.n).fillna(0)
        inv_down = cogs.cumsum()
        billed = inv_up
        paid = billed.shift(self.AP_COLLECTION.n).fillna(0) # shift one extra; booked on last day of month
        
        base.loc['Inventory', :] = inv_up.values -  inv_down.values
        base.loc['Accounts Payable', :] = billed.values - paid.values  + base.loc['Accounts Payable'].iloc[0].values

        # Capex and Amort EASY #
        capex = -self.capexsched.sum().values
        base.loc['Purchases of PPE', :] = capex
        base.loc['Capital Assets', :] = capex.cumsum() - self.amortsched.cumsum().values[:capex.size]

        # Capital Accounts #
        stock_sales = self.capital.groupby('Account').sum().loc['Stock Sale Proceeds', :].values
        base.loc['Stock Sale Proceeds', :] = stock_sales
        base.loc['Equity Capital', :] = stock_sales.cumsum() + base.loc['Equity Capital', :].iloc[0].values

        #### FIRST CHANGE TO CASH ####
        toe = self._isprox.loc['Total Operating Expenses', :].values
        base.loc['Shadow Cash', :] = collected.values - paid.values - capex.cumsum() + stock_sales.cumsum() - toe.cumsum() + base.loc['Shadow Cash', :].iloc[0].values

        # Interest Impact #
        interest = np.zeros(self.projrng.size)
        red_cash_mask = base.loc['Shadow Cash'].iloc[0] < 0
        interest[red_cash_mask] = -base.loc['Shadow Cash'].values[0][red_cash_mask] * self.INT_RATE
        base.loc['Interest', :] = interest
        
        ebt = self._isprox.loc['EBIT', :].values - interest
        self._isprox.loc['EBT', :] = ebt

        ### TAXES ###
        # First, find taxes and find when accumulated taxes are positive and negative
        tax = ebt*self.TAX_RATE   
        cumtax = tax.cumsum() - base.loc['Tax Carry-Forward', self.projrng[0]].values[0] # cumulative tax must incorporate tax carryforwards
        red_tax_mask = cumtax < 0
        black_tax_mask = cumtax >= 0
        
        # Second, assign taxes and cumulative taxes to carry-forward; 
        # carry-forward is 0 when cumtax >= 0
        base.loc['Tax Carry-Forward', base.columns[red_tax_mask]] = -cumtax[red_tax_mask]
        base.loc['Taxes', :] = tax
        
        # Second, determine cash portion of txes
        # When cumulative taxes > 0, cash taxes == taxes
        cashtax = np.zeros(self.projrng.size)
        cashtax[black_tax_mask] = tax[black_tax_mask]

        # There are some periods where carry-fwd > 0, but cash taxes are still paid
        # this occurs when cash taxes_t > carry-fwd_t-1
        # So must find periods when carry-fwd_t-1 > 0 AND cash taxes_t > carry-fwd_t-1
        black_fwd_mask = base.loc['Tax Carry-Forward'].iloc[0].shift(1).fillna(0) > 0
        tax_gt_fwd_mask = base.loc['Taxes'].iloc[0].fillna(0) - base.loc['Tax Carry-Forward'].iloc[0].shift(1).fillna(0) > 0
        partial_cash_tax_mask = black_fwd_mask & tax_gt_fwd_mask

        # Find cashtaxes for these instances and assign
        partial_fwd = base.loc['Tax Carry-Forward'].iloc[0].shift(1).fillna(0)[partial_cash_tax_mask]
        partial_tax = base.loc['Taxes'].iloc[0][partial_cash_tax_mask]
        cashtax[partial_cash_tax_mask] = partial_tax - partial_fwd
        base.loc['Cash Taxes', :] = cashtax
        base.loc['Non-Cash Taxes', :] = tax - cashtax
        
        ni = ebt - tax
        base.loc['Net Income', :] = ni
        base.loc['Retained Earnings', :] = ni.cumsum() + base.loc['Retained Earnings'].iloc[0].values

        base.loc['Shadow Cash'] -=  interest.cumsum() + cashtax.cumsum() 

        base.loc['Cash', :] = base.loc['Shadow Cash', :].where(base.loc['Shadow Cash', :] >= 0, 0).values
        base.loc['Indebtedness', :] = -base.loc['Shadow Cash', :].where(base.loc['Shadow Cash', :] <= 0, 0).values

        return base

    def update(self, metrics=[], ops=[], refresh=False, hide=True):
        super().update(refresh=refresh)

        for acct in self.ACCOUNTS.tabs:
            self._statement = acct.append(self._statement)

        metrics += ['wocap', 'cap', 'd2cap', 'crat']
        for metric in metrics:
            metric_acct = self.ACCOUNTS.get(metric)

            if metric_acct is None:
                raise TypeError(f'{metric} is not a valid Account name. Did you include it in list?')

            self._statement = metric_acct.append(self._statement)            

        if ops:
            ops = [ops] if isinstance(ops, str) else ops
            for op in ops:
                self._statement = getattr(self, op)(self._statement)

        self._statement = self._statement.fillna(0).sort_index()
        
        if hide:
            show_mask = np.array([self.ACCOUNTS.get(acct).show for acct in self.stat.index.get_level_values('Account')])
            self._statement = self._statement[show_mask]
        
        return self

class EQStat(BaseBSheet):
    FULL_NAME = 'Statement of Changes in Equity'
    ACCOUNTS = EQ_ACCOUNTS

    first = {'first': ['Retained Earnings - Start', 'Equity Capital - Start']}
    last = {'last': [
        'Retained Earnings', 'Equity Capital', 
        "Shareholders' Equity"
    ]}
    sums = {'sum': ['Net Income', 'Stock Sale Proceeds', ]}
    aggfuncs = [first, last, sums] 

    class Index(IndexMix):
        Account = EQ_ACCOUNTS.names.tolist()

    def __init__(self, bsheet):
        super().__init__()
        self._projrng = bsheet._projrng
        
        self._index = self.Index(self)
        self._bsheet = bsheet
        self._base = self._make_base()
        self.update(refresh=True)

    def _make_base(self):
        idx_from_bsheet = self._bsheet.stat.index.get_level_values('Account').intersection(self.index.levels['Account'])
        base = self._bsheet.stat.loc[idx_from_bsheet]
        base = self._refresh_index(base, idx_from_bsheet)
        
        return base
        
    def _refresh_index(self, base, idx_from_bsheet):
        acctidx = pd.Categorical(idx_from_bsheet, categories=self.index.levels['Account'], ordered=True)
        index_as_frame = pd.DataFrame({'Account': acctidx})
        base.index = pd.MultiIndex.from_frame(index_as_frame)
        
        return base

    def update(self, metrics=[], ops=[], refresh=False, hide=True):
        super().update(refresh=refresh)

        for acct_name in self.ZERO_ROWS:
            if acct_name not in self._statement.index.get_level_values('Account'):
                self._add_zero_row(acct_name, self._statement)

        for acct in self.ACCOUNTS.tabs:
            self._statement = acct.append(self._statement)

        self._statement = self._statement.fillna(0).sort_index()

        return self

class CFlow(EQStat):
    FULL_NAME = 'Cash Flow Statement'
    ACCOUNTS = CFLOW_ACCOUNTS
    ZERO_ROWS = ['Proceeds from Sale', 'Proceeds of LTD', 'Repayment of LTD', 'Dividends Paid']

    class Index(IndexMix):
        Account = CFLOW_ACCOUNTS.names.tolist()

    @property
    def aggfuncs(self):
        first = {'first': ['Cash - Start']}
        last = {'last': [
            'Cash - End'
        ]}
        sums = {'sum': self.stat.index.get_level_values('Account').difference(['Cash - Start', 'Cash - End'])}
        return [first, last, sums] 

    def _make_base(self):
        idx_from_bsheet = self._bsheet.stat.index.get_level_values('Account').intersection(self.index.levels['Account'])
        idx_from_isprox = self._bsheet._isprox.index.get_level_values('Account').intersection(self.index.levels['Account'])
        idx_from_isprox = idx_from_isprox.difference(idx_from_bsheet)
        idx_already = idx_from_bsheet.append(idx_from_isprox)
        base = self._bsheet.stat.loc[idx_from_bsheet]
        base = base.append(self._bsheet._isprox.loc[idx_from_isprox])
        base = self._refresh_index(base, idx_already)
        base.loc['Purchases of PPE', :] *= -1
        
        base.loc['Cash - End', :] = 0

        return base

    def update(self, metrics=[], ops=[], refresh=False, hide=True):
        if refresh:
            self._statement = self._base.copy()

        for acct_name in self.ZERO_ROWS:
            if acct_name not in self._statement.index.get_level_values('Account'):
                self._add_zero_row(acct_name, self._statement)

        for acct in self.ACCOUNTS.tabs:
            self._statement = acct.append(self._statement)

        acct = self.ACCOUNTS.get('Cash - Start')
        self._statement = self._statement.drop('Cash - Start', level='Account')
        self._statement = acct.append(self._statement)

        self._statement = self._statement.fillna(0).sort_index()

        if hide:
            show_mask = np.array([self.ACCOUNTS.get(acct).show for acct in self.stat.index.get_level_values('Account')])
            self._statement = self._statement[show_mask]
        
        return self

### MAIN INTERFACE ###
class FinStat(BaseStat):
    INT_RATE = .04 / 12
    TAX_RATE = .35
    STAT_NAMES = ['IStat', 'BSheet', 'EQStat', 'CFlow']

    @set_assumps
    def __init__(self, 
        revenue, wages, headcount, opex, capex, capital,
        assumps=[],
        opening=None,
        make_plots=False,
        plot_root=''
        ):

        print ('Generating Income Statement', end='\r')
        self._istat = IStat(revenue, wages, headcount, opex, capex, assumps=assumps)
        print ('Generating Balance Sheet      ', end='\r')                
        self._bsheet = BSheet(self._istat, capital, assumps=assumps, opening=opening)
        print ('Generating Equity Statement', end='\r')        
        self._eqstat = EQStat(self._bsheet)
        print ('Generating Cash Flow Statement')
        self._cflow = CFlow(self._bsheet)
        print ('Done.                        ')
        
        if make_plots:
            self._plots = Plots(self, root=plot_root)
            self._plots.make_all()

    @property
    def _supporting(self):        
        return [
            self.istat.rev, self.istat.wages, self.istat.amort
        ]

    def supporting(self, as_dict=False):
        if as_dict:
            return {stat.FULL_NAME: stat for stat in self._supporting}
        else:
            return self._supporting
    
    @property
    def istat(self):
        return self._istat

    @property
    def bsheet(self):
        return self._bsheet
        
    @property
    def eqstat(self):
        return self._eqstat
    
    @property
    def cflow(self):
        return self._cflow

    @property
    def _stats(self):
        return [self._istat, self._bsheet, self._eqstat, self._cflow]
    
    def stats(self, as_dict=False):
        if as_dict:
            return {k: v for k, v in zip(self.STAT_NAMES, self._stats)}
        else:
            return  self._stats
    
    @property
    def annuals(self):
        return {name: stat.annually() for name, stat in self.stats(as_dict=True).items()}

    @property
    def quarterlies(self):
        return {name: stat.quarterly() for name, stat in self.stats(as_dict=True).items()}

    @property
    def assumptions(self):
        return pd.DataFrame.from_dict(
            {
                'Income Statement': ['', ''],
                'Interest Rate': [self.INT_RATE * 12, 'Annual'],
                'Income Tax Rate': [self.TAX_RATE, 'Annual'],
                'Balance Sheet': ['', ''],
                'AR Terms': [self.bsheet.AR_COLLECTION.n, 'Months'],
                'AP Terms': [self.bsheet.AP_COLLECTION.n, 'Months'],
                'Inventory Lead Time': [self.bsheet.INVENTORY_LEAD.n, 'Months'],
                'Amortization Schedule': [AmortSched.AMORT_YEARS, 'Years'],
                'Headcount Items': ['', ''],
                'Payroll Taxes': [WageStat.PAYROLL_TAXES, 'Annual'],
                'Wage Inflation': [WageStat.WAGE_INFLATION, 'Annual'],
                'Benefits': [WageStat.ANNUAL_BENEFITS, 'per person'],
                'Payroll Fees': [WageStat.ANNUAL_PAYROLL_FEES, 'per person'],
                'Unemployment Fees': [WageStat.ANNUAL_UNEMPLOYMENT, 'per person'],
                "Workers' Compensation": [WageStat.ANNUAL_WORKERS_COMP, 'Annual (per salaries)'],
                "401K Contributions": [WageStat.ANNUAL_401K, 'Annual (per salaries)']
            }, orient='index', columns=['Factor', 'Units / Period']
        )

    @property
    def plots(self):
        return self._plots



import os
import numpy as np
import pandas as pd
import itertools as it

import xlwings as xw

from finstat.style import tabs
from finstat.finstat import FinStat

IMG_KWARGS = {
    'istat': {
        'left': 0,
        'top': 0,
    },
    'revpie': {
        'left': 800,
        'top': 0
    },
    'cap_struct': {
        'left': 0,
        'top': 600
    },
    'wocap': {
        'left': 800,
        'top': 600
    },
}

@xw.sub
def main():
    wb = xw.Book.caller()
    args = compile_sheets(wb)
    cash = 1620209
    ar = 89511
    # inventory = 720712
    capass = 76380
    ap = 221100 + 60554 + 20023 + 223133 + 58034
    nols = 15380552
    taxcrryfwd = nols * .35
    plug = 595173
    eq = 14756919 + 1447777 + 2000000 + 1441094 - plug + taxcrryfwd - ar
    re = -16417861 - 1505880
    opening = pd.DataFrame.from_dict({
        'Shadow Cash': cash,
        'Tax Carry-Forward': taxcrryfwd,
        'Accounts Payable': ap,
        'Equity Capital': eq,
        'Retained Earnings': re,
    }, orient='index')
    fstat = FinStat(*args, opening=opening, make_plots=True)
    print ('Add Spacers')
    add_spacers(wb)
    print ('Add Summary')
    add_summary(wb, fstat)
    print ('Expense Analysis')
    add_exp_anal(wb, fstat)
    print ('Write Data')
    write_data(wb, fstat)
    print ('Write Expense Schedules')
    add_exp_schedule(wb, fstat)
    print ('Color Tabs')
    color_tabs(wb)
    print ('Complete')

def compile_sheets(wb):
    pipeline = wb.sheets['Pipeline'].range('A1:BA200').options(pd.DataFrame, index=False, header=False, numbers=float).value
    pipeline = pipeline.fillna(np.nan)

    assumps = wb.sheets['Assumptions'].range('A1').expand().options(pd.DataFrame).value
    assumps = assumps.fillna(np.nan)

    wages = wb.sheets['Wages - Raw'].range('A1').expand().options(pd.DataFrame, index=False, header=False).value
    wages = wages.fillna(np.nan)

    hcraw = wb.sheets['Headcount - Raw'].range('A1').expand().options(pd.DataFrame, index=False, header=False).value

    opex = wb.sheets['Opex'].range('A1:R100').options(pd.DataFrame, index=False, header=False).value
    idxs = np.argwhere(opex.isna().all(axis=1).values).flatten()
    last = idxs[(idxs - np.roll(idxs, 1)) == 1][0]
    opex = opex.iloc[:last-1]
    opex = opex.loc[:, ~opex.isna().all(axis=0)]
    opex = opex.fillna(np.nan)

    capex = wb.sheets['Capex'].range('A1').expand().options(pd.DataFrame, index=False, header=False).value
    capex = capex.fillna(np.nan)

    capital = wb.sheets['Capital'].range('A1').expand().options(pd.DataFrame, index=False).value
    capital = capital.fillna(np.nan)

    return pipeline, wages, hcraw, opex, capex, capital, assumps

# Add Spacers if Required
def add_spacers(wb):
    for name in tabs.spacers:
        if name not in [s.name for s in wb.sheets]:
            after = tabs.before_after.loc[name, 'after']
            wb.sheets.add(name, after=after)

def add_exp_anal(wb, fstat):
    if 'Expenses' not in [s.name for s in wb.sheets]:
        exp = wb.sheets.add('Expenses', after=tabs.spacers[1])
    else:
        exp = wb.sheets['Expenses']

    exp.range('A2').expand('table').value = fstat.istat.opex.top10()

def add_summary(wb, fstat):
    if 'Charts' not in [s.name for s in wb.sheets]:
        summ = wb.sheets.add('Charts', after=tabs.spacers[1])
    else:
        summ = wb.sheets['Charts']

    for plot, fig in fstat.plots.plots.items():
        kws = IMG_KWARGS[plot]
        summ.pictures.add(fig, name=plot, update=True, **kws)

def write_data(wb, fstat):
    for period, stat in it.product(['A', 'Q'], fstat.STAT_NAMES):
        statobj = getattr(fstat, stat.lower())
        
        update_kws = {'hide': True}
        if stat.lower() == 'istat':
            update_kws = {'ops': ['_merge_benefits'], **update_kws}

        if stat.lower() == 'istat':
            frame = fstat.istat.update(refresh=True, **update_kws).resample(period, accts=True, str_cols=False, clip_rng=fstat.istat.stat.columns.tolist())
        else:
            frame = statobj.update(refresh=True, **update_kws).resample(period, accts=True, str_cols=False, clip_rng=fstat.istat.stat.columns.tolist())

        strf = '%Y' if period == 'A-DEC' else '%b-%y'
        frame.columns = frame.columns.strftime(strf)
        wb.sheets[f'{stat} - {period}'][:frame.shape[1] + 1 , :frame.shape[0] + 1].options(index=True, header=True).value = frame

def add_exp_schedule(wb, fstat):
    if 'Expense Schedule - A' not in [s.name for s in wb.sheets]:
        exp = wb.sheets.add('Expense Schedule - A', before=tabs.spacers[-2])
    else:
        exp = wb.sheets['Expense Schedule - A']

    annual = fstat.istat._opex.exp_schedule().isutils.annually()
    annual.columns = annual.columns.strftime('%b-%y')

    exp.range('A1').expand('table').value = annual

    if 'Expense Schedule - Q' not in [s.name for s in wb.sheets]:
        exp = wb.sheets.add('Expense Schedule - Q', before=tabs.spacers[-1])
    else:
        exp = wb.sheets['Expense Schedule - Q']

    quarterly = fstat.istat._opex.exp_schedule()
    quarterly.columns = quarterly.columns.strftime('%b-%y')

    exp.range('A1').expand('table').value = quarterly

def color_tabs(wb):
    for sheet in wb.sheets:
        tabs.set_color(sheet)
        


import os
from pathlib import Path
import numpy as np
import pandas as pd
import functools as ft

import matplotlib.pyplot as plt
import matplotlib.ticker as mticks
import matplotlib.image as mimage

from finstat.style import grc

def returner(func):
    @ft.wraps(func)
    def wrapper(self, *args, return_fig=False, **kwargs):
        fig, fname = func(self, *args, **kwargs)

        plt.savefig(self.IMG_PATH / fname, bbox_inches='tight')
        
        if return_fig:
            return fig

    return wrapper

class Plots:
    FREQKEY = {'Q-DEC': 'Quarterly', 'A-DEC': 'Annual', 'M': 'Monthly'}
    names = ['istat', 'cap_struct', 'wocap', 'revpie']

    def __init__(self, fstat, figsize=(12,8), root=''):
        import builtins

        self.ROOT_DIR = Path(root) if root else Path(builtins.FORECAST_ROOT)
        self.fstat = fstat
        self.figsize = figsize

    @property
    def LOGO_PATH(self):
        return self.ROOT_DIR / 'logo.png'
    
    @property
    def IMG_PATH(self):
        return self.ROOT_DIR / 'charts'

    def _insert_logo(self, ax, bbox):
        im = mimage.imread(self.LOGO_PATH)
        imax = ax.inset_axes(bbox, transform=ax.transAxes, zorder=10)
        imax.axis('off')
        imax.imshow(im, aspect='auto', extent=(0.4, 0.6, .5, .7), zorder=10)

    def make_all(self, return_figs=True):
        self._plots = {}
        for name in self.names:
            self._plots[name] = getattr(self, name)(return_fig=return_figs)

        return self

    @property
    def plots(self):
        if not hasattr(self, '_plots'):
            raise AttributeError('You must call `make_all` first')
        else:
            return self._plots

    @returner
    def istat(self, stat=None, return_fig=False):
        self.fstat.istat.update(refresh=True)
        stat = self.fstat.istat.utils.resample('Q', accts=True, aggfunc='sum', clip_rng=True) if stat == None else stat

        fig, ax = plt.subplots(figsize=self.figsize)

        xindex = np.arange(stat.columns.size)
        ax.plot(stat.loc['Revenue'].values, lw=4, c=grc.blue.data, label='Revenue')
        ax.plot(stat.loc['Gross Profit'].values, lw=4, c=grc.lblue.data, label='Gross Profit')
        ax.plot(stat.loc['Net Income'].values, lw=4, c=grc.green.data, label='Net Income')

        ax.set_xticks(np.arange(stat.columns.size))
        ax.set_xticklabels([stat.columns[i].strftime('%b\n%Y') if not i%4 else None for i in range(xindex.size)])

        ax.yaxis.set_major_formatter('${x:,.0f}')

        ax.spines['top'].set_visible(False)
        ax.spines['right'].set_visible(False)

        ax.legend(title='\n')
        self._insert_logo(ax, [.025, .93, .1, .045])

        freq = stat.T.resample('Q').sum().T.columns.freqstr
        ax.set_title(f'Income Statement\n{self.FREQKEY[freq]}', y=1, size=32)

        return fig, 'istat.jpg'

    @returner
    def cap_struct(self, bsheet=None, show_d2cap=True, return_fig=False):
        bsheet = self.fstat.bsheet.utils.quarterly(clip_rng=True) if bsheet == None else bsheet

        fig, ax = plt.subplots(figsize=self.figsize)

        xindex = np.arange(bsheet.columns.size)
        bar1 = ax.bar(
            xindex, 
            bsheet.loc["Shareholders' Equity"].values[0], 
            width=.9, 
            color=grc.green.data,
            label='Equity     ',
            tick_label=bsheet.columns.strftime('%Y-%m'),
        )
        bar2 = ax.bar(
            xindex,
            bsheet.loc['Indebtedness'].values[0],
            width=.9,
            color=grc.blue.data,    
            bottom=bsheet.loc["Shareholders' Equity"].values[0],
            label='Debt    '
        )

        ax.set_xticklabels([bsheet.columns[i].strftime('%b\n%Y') if not (i + 1) % 4 else None for i in range(xindex.size)])
        ax.yaxis.set_major_formatter('{x:,.0f}')

        ax.spines['top'].set_visible(False)
        ax.spines['right'].set_visible(False)

        if show_d2cap:
            twinx = ax.twinx()
            dcline, = twinx.plot(xindex, bsheet.loc['Debt-to-Capital'].values[0], lw=4, color='r', label='Debt-to-Capital')

            twinx.yaxis.set_major_formatter('{x:.0%}')
            twinx.spines['top'].set_visible(False)
            twinx.legend([bar1, bar2, dcline], ['Equity', 'Debt', 'Debt-to-Cap'], loc='upper left', framealpha=1, title='\n')
            self._insert_logo(twinx, [.02825, .93, .1, .045])
        else:
            ax.legend(title='\n')
            self._insert_logo(ax, [.02, .93, .1, .045])

        ax.set_title('Capital Structure \nMonthly', y=1, size=32)

        return fig, 'capstruct.jpg'

    @returner
    def wocap(self, bsheet=None, show_crat=True, return_fig=False):
        bsheet = self.fstat.bsheet.utils.quarterly(clip_rng=True) if bsheet == None else bsheet

        fig, ax = plt.subplots(figsize=self.figsize)

        xindex = np.arange(bsheet.columns.size)
        w = 0.35

        rects1 = ax.bar(xindex - w/2, bsheet.loc['Current Assets'].values[0], w, color=grc.blue, label='Current Assets')
        rects2 = ax.bar(xindex + w/2, bsheet.loc['Current Liabilities'].values[0], w, color=grc.green, label='Current Liabilities')

        ax.set_xticks(np.arange(bsheet.columns.size))
        ax.set_xticklabels([bsheet.columns[i].strftime('%b\n%Y') if not i%4 else None for i in range(xindex.size)])

        ax.yaxis.set_major_formatter(mticks.FuncFormatter(lambda x, p: f'${x/1e6:,.0f}M'))

        ax.spines['top'].set_visible(False)
        ax.spines['right'].set_visible(False)

        if show_crat:
            twinx = ax.twinx()
            cratline, = twinx.plot(xindex[4:], bsheet.loc['Current Ratio'].values[0][4:], lw=4, color='r', label='Current Ratio')

            twinx.spines['top'].set_visible(False)
            twinx.yaxis.set_major_formatter('{x:.1f}x')

            twinx.legend([rects1, rects2, cratline], ['Current Assets', 'Current Liabilities', 'Current Ratio'], loc='upper left', title='\n')
            self._insert_logo(twinx, [.0475, .93, .1, .045])
        else:
            ax.legend(title='\n')
            self._insert_logo(ax, [.02, .93, .1, .045])

        freq = bsheet.columns.freqstr    
        ax.set_title(f'Working Capital\n{self.FREQKEY[freq]}', y=1, size=32)

        return fig, 'wocap.jpg'

    @returner
    def revpie(self, istat=None, return_fig=False):
        # Pie chart, where the slices will be ordered and plotted counter-clockwise:
        istat = self.fstat.istat if istat is None else istat

        rev_by_sgmt = istat.rev.utils.resample('A', sgmts=True).loc[pd.IndexSlice[:, 'Revenue'], :]
        rev_by_sgmt = rev_by_sgmt.reset_index().set_index('Segment').drop(columns='Account')
        rev_by_sgmt = rev_by_sgmt / rev_by_sgmt.sum(axis=0)

        labels = rev_by_sgmt.index.values
        sizes1 = rev_by_sgmt.loc[:, rev_by_sgmt.columns[0]]
        sizes2 = rev_by_sgmt.loc[:, rev_by_sgmt.columns[1]]

        fig, (ax1, ax2) = plt.subplots(1,2, figsize=(12,8))
        ax1.pie(
            sizes1, labels=labels, autopct='%1.0f%%', colors=[grc.blue, grc.green, grc.lblue],
            startangle=90, textprops={'size': 14}
        )
        ax2.pie(
            sizes2, labels=labels, autopct='%1.0f%%', colors=[grc.blue, grc.green, grc.lblue],
            startangle=90, textprops={'size': 14}
        )

        ax1.axis('equal')  # Equal aspect ratio ensures that pie is drawn as a circle.
        ax2.axis('equal')

        ax1.set_title('2022', y=.85, fontsize=24)
        ax2.set_title('2023', y=.85, fontsize=24)
        plt.suptitle('Revenue by Region', y=.9, fontsize=32)
        
        return fig, 'revpie.jpg'

    @returner
    def multibar(self, istat=None, values=[], style='', return_fig=False):
        istat = self.fstat.istat if istat is None else istat

        opex_accts = istat._wages.ACCOUNTS + istat._opex.ACCOUNTS
        segopex = istat.annually().loc[opex_accts.names, :].groupby('Segment').sum()

        if style == 'per_rev':
            segopex /= istat.annually().loc['Total Revenue'].values
        elif style == 'hc':
            segopex = istat.wages._headcount.T.resample('A-DEC').last().T.groupby('Segment').sum()

        mid = segopex.shape[0] / 2
        spreads = np.arange(-np.floor(mid), np.ceil(mid))

        x = np.arange(segopex.shape[1])  # the label locations
        w = 0.15  # the width of the bars

        fig, ax = plt.subplots()
        rects = {}
        colors = grc.colors()
        for i, (sgmt, row) in enumerate(segopex.iloc[::-1].iterrows()):
            rects[sgmt] = ax.bar(x - w*spreads[i], row, w, label=sgmt, color=colors[i].data, zorder=5)

        # Add some text for labels, title and custom x-axis tick labels, etc.
        ax.set_xticks(x)
        ax.set_xticklabels(segopex.columns)

        import matplotlib.ticker as mticks
        @mticks.FuncFormatter
        def mil_fmt(x, pos):
            return f'${x/1e6:,.0f}M'

        if style == 'per_rev':
            ax.yaxis.set_major_formatter(mticks.PercentFormatter(1, decimals=0))
            title = 'Opex by Segment \n(as % of Total Revenue)'
        elif style == 'hc':
            title = 'Headcount by Segment'       
        else:
            title = 'Opex by Segment \n(including Salaries & Bonuses)'
            ax.yaxis.set_major_formatter(mil_fmt)

        ax.tick_params('y', width=0)

        ax.set_title(title)
        ax.legend(framealpha=1)

        for spine in ax.spines:
            ax.spines[spine].set_visible(False)

        ax.grid(axis='y', zorder=1, alpha=.5)

        filename = f'multi_{style}.jpeg'
        return fig, filename
